"""Additive models."""

from __future__ import annotations

from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
from jax.scipy import stats

from elisa.models.model import (
    AnaIntAdditive,
    NumIntAdditive,
    ParamConfig,
)

if TYPE_CHECKING:
    from elisa.util.typing import CompEval, JAXArray, NameValMapping

__all__ = [
    'Band',
    'BandEp',
    'BentPL',
    'Blackbody',
    'BlackbodyRad',
    'BrokenPL',
    'Compt',
    'CutoffPL',
    'Gauss',
    'Lorentz',
    'OTTB',
    'OTTS',
    'PLEnFlux',
    'PLPhFlux',
    'PowerLaw',
    'SmoothlyBrokenPL',
]


def _powerlaw_integral(egrid: JAXArray, alpha: JAXArray) -> JAXArray:
    cond = jnp.full(len(egrid), jnp.not_equal(alpha, 1.0))

    one_minus_alpha = jnp.where(cond, 1.0 - alpha, 1.0)
    f1 = jnp.power(egrid, one_minus_alpha) / one_minus_alpha
    f1 = f1[1:] - f1[:-1]

    f2 = jnp.log(egrid)
    f2 = f2[1:] - f2[:-1]

    return jnp.where(cond[:-1], f1, f2)


class Band(NumIntAdditive):
    r"""Gamma-ray burst continuum developed by Band et al. (1993) [1]_.

    .. math::
        N(E) = K
        \begin{cases}
        \bigl(\frac{E}{E_0}\bigr)^\alpha
            \exp\bigl(-\frac{E}{E_\mathrm{c}}\bigr),
            &\text{if } E < (\alpha-\beta) E_\mathrm{c},
        \\\\
        \bigl(\frac{E}{E_0}\bigr)^\beta \exp(\beta-\alpha)
            \left[
                \frac{(\alpha-\beta)E_\mathrm{c}}{E_0}
            \right]^{\alpha-\beta},
            &\text{otherwise,}
        \end{cases}

    where :math:`E_0` is the reference energy fixed at 100 keV.

    Parameters
    ----------
    alpha : Parameter, optional
        The low-energy power law index :math:`\alpha`, dimensionless.
    beta : Parameter, optional
        The high-energy power law index :math:`\beta`, dimensionless.
    Ec : Parameter, optional
        The characteristic energy :math:`E_\mathrm{c}`, in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.

    References
    ----------
    .. [1] `Band, D., et al. 1993, ApJ, 413, 281
           <https://adsabs.harvard.edu/full/1993ApJ...413..281B>`__
    """

    _config = (
        ParamConfig('alpha', r'\alpha', '', -1.0, -10.0, 5.0),
        ParamConfig('beta', r'\beta', '', -2.0, -10.0, 10.0),
        ParamConfig('Ec', r'E_\mathrm{c}', 'keV', 300.0, 10.0, 1e4),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        alpha = params['alpha']
        beta = params['beta']
        Ec = params['Ec']
        K = params['K']

        e0 = 100.0

        # workaround for beta > alpha, as in XSPEC
        amb_ = alpha - beta
        inv_Ec = 1.0 / Ec
        amb = jnp.where(jnp.less(amb_, inv_Ec), inv_Ec, amb_)
        Ebreak = Ec * amb

        log = jnp.where(
            jnp.less(egrid, Ebreak),
            alpha * jnp.log(egrid / e0) - egrid / Ec,
            amb * jnp.log(amb * Ec / e0) - amb + beta * jnp.log(egrid / e0),
        )

        return K * jnp.exp(log)


class BandEp(NumIntAdditive):
    r"""Gamma-ray burst continuum developed by Band et al. (1993) [1]_,
    parametrized by the peak of :math:`\nu F_\nu`.

    .. math::
        N(E) = K
        \begin{cases}
        \bigl(\frac{E}{E_0}\bigr)^\alpha
            \exp\left[-\frac{(2+\alpha)E}{E_\mathrm{p}}\right],
            &\text{if } E < \frac{(\alpha-\beta)E_\mathrm{p}}{2+\alpha},
        \\\\
        \bigl(\frac{E}{E_0}\bigr)^\beta \exp(\beta-\alpha)
            \left[
                \frac{(\alpha-\beta)E_\mathrm{p}}{(2+\alpha)E_0}
            \right]^{\alpha-\beta},
            &\text{otherwise},
        \end{cases}

    where :math:`E_0` is the reference energy fixed at 100 keV.

    .. warning::
        This model requires the low-energy power law index :math:`\alpha`
        to be greater than -2.0.

    Parameters
    ----------
    alpha : Parameter, optional
        The low-energy power law index :math:`\alpha`, dimensionless.
    beta : Parameter, optional
        The high-energy power law index :math:`\beta`, dimensionless.
    Ep : Parameter, optional
        The peak energy :math:`E_\mathrm{p}` of :math:`\nu F_\nu`,
        in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.

    References
    ----------
    .. [1] `Band, D., et al. 1993, ApJ, 413, 281
           <https://adsabs.harvard.edu/full/1993ApJ...413..281B>`__
    """

    _config = (
        ParamConfig('alpha', r'\alpha', '', -1.0, -1.99, 10.0),
        ParamConfig('beta', r'\beta', '', -2.0, -10.0, 10.0),
        ParamConfig('Ep', r'E_\mathrm{p}', 'keV', 300.0, 10.0, 1e4),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        alpha = params['alpha']
        beta = params['beta']
        Ep = params['Ep']
        K = params['K']

        e0 = 100.0

        Ec = Ep / (2.0 + alpha)
        Ebreak = (alpha - beta) * Ec

        # workaround for beta > alpha, as in XSPEC
        amb_ = alpha - beta
        inv_Ec = 1.0 / Ec
        amb = jnp.where(jnp.less(amb_, inv_Ec), inv_Ec, amb_)

        log = jnp.where(
            jnp.less(egrid, Ebreak),
            alpha * jnp.log(egrid / e0) - egrid / Ec,
            amb * jnp.log(amb * Ec / e0) - amb + beta * jnp.log(egrid / e0),
        )

        return K * jnp.exp(log)


class BentPL(NumIntAdditive):
    r"""Bent power law.

    .. math::
        N(E) = 2K \left[
            1 + \left(\frac{E}{E_\mathrm{b}}\right)^\alpha
        \right]^{-1}

    Parameters
    ----------
    alpha : Parameter, optional
        The power law photon index :math:`\alpha`, dimensionless.
    Eb : Parameter, optional
        The break energy :math:`E_\mathrm{b}`, in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('alpha', 'alpha', '', 1.5, -3.0, 10.0),
        ParamConfig('Eb', r'E_\mathrm{b}', 'keV', 5.0, 1e-2, 1e6),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        alpha = params['alpha']
        Eb = params['Eb']
        K = params['K']
        return K * 2.0 / (1.0 + jnp.power(egrid / Eb, alpha))


class Blackbody(NumIntAdditive):
    r"""Blackbody function.

    .. math::
        N(E) = \frac{C K E^2}{(kT)^4 [\exp(E/kT)-1]},

    where :math:`C=8.0525` ph.

    Parameters
    ----------
    kT : Parameter, optional
        The temperature :math:`kT`, in units of keV.
    K : Parameter, optional
        The amplitude :math:`K = L_{39}/D_{10}^2`, where :math:`L_{39}` is the
        source luminosity in units of 10³⁹ erg s⁻¹ and :math:`D_{10}` is the
        distance to the source in units of 10 kpc.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('kT', 'kT', 'keV', 3.0, 1e-4, 200.0),
        ParamConfig('K', 'K', '10^37 erg s^-1 kpc^-2', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        kT = params['kT']
        K = params['K']
        x = egrid / kT
        tmp = 8.0525 * K * egrid / (kT * kT * kT)
        x_ = jnp.where(
            jnp.greater_equal(x, 50.0),
            1.0,  # avoid exponential overflow
            x,
        )

        return jnp.where(
            jnp.less_equal(x, 1e-4),
            tmp,
            jnp.where(
                jnp.greater_equal(x, 50.0),
                0.0,  # avoid exponential overflow
                tmp * x / jnp.expm1(x_),
            ),
        )
        # return 8.0525 * K * e*e / (kT*kT*kT*kT * jnp.expm1(e / kT))


class BlackbodyRad(NumIntAdditive):
    r"""Blackbody function with normalization proportional to the surface area.

    .. math::
        N(E) = \frac{C K E^2}{\exp(E/kT)-1},

    where :math:`C=1.0344 \times 10^{-3}` ph cm⁻² s⁻¹ keV⁻³.

    Parameters
    ----------
    kT : Parameter, optional
        The temperature :math:`kT`, in units of keV.
    K : Parameter, optional
        The amplitude :math:`K = R_\mathrm{km}^2/D_{10}^2`, where
        :math:`R_\mathrm{km}` is the source radius in km and :math:`D_{10}` is
        the distance to the source in units of 10 kpc.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('kT', 'kT', 'keV', 3.0, 1e-4, 200.0),
        ParamConfig('K', 'K', '', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        kT = params['kT']
        K = params['K']

        x = egrid / kT
        tmp = 1.0344e-3 * K * egrid
        x_ = jnp.where(
            jnp.greater_equal(x, 50.0),
            1.0,  # avoid exponential overflow
            x,
        )

        return jnp.where(
            jnp.less_equal(x, 1e-4),
            tmp * kT,
            jnp.where(
                jnp.greater_equal(x, 50.0),
                0.0,  # avoid exponential overflow
                tmp * egrid / jnp.expm1(x_),
            ),
        )
        # return 1.0344e-3 * K * e*e / jnp.expm1(e / kT)


class BrokenPL(AnaIntAdditive):
    r"""Broken power law.

    .. math::
        N(E) = K
        \begin{cases}
        \left(\frac{E}{E_\mathrm{b}}\right)^{-\alpha_1},
            &\text{if } E \le E_\mathrm{b},
        \\\\
        \left(\frac{E}{E_\mathrm{b}}\right)^{-\alpha_2},
            &\text{otherwise}.
        \end{cases}

    Parameters
    ----------
    alpha1 : Parameter, optional
        The low-energy power law photon index :math:`\alpha_1`, dimensionless.
    alpha2 : Parameter, optional
        The high-energy power law photon index :math:`\alpha_2`, dimensionless.
    Eb : Parameter, optional
        The break energy :math:`E_\mathrm{b}`, in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('alpha1', 'alpha_1', '', 1.0, -3.0, 10.0),
        ParamConfig('Eb', r'E_\mathrm{b}', 'keV', 5.0, 1e-2, 1e6),
        ParamConfig('alpha2', 'alpha_2', '', 2.0, -3.0, 10.0),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        alpha1 = params['alpha1']
        Eb = params['Eb']
        alpha2 = params['alpha2']
        K = params['K']
        mask = egrid[:-1] <= Eb
        egrid_ = egrid / Eb
        p1 = _powerlaw_integral(egrid_, alpha1)
        p2 = _powerlaw_integral(egrid_, alpha2)
        f = jnp.where(mask, p1, p2)
        idx = jnp.flatnonzero(mask, size=egrid.size - 1)[-1]
        pb1 = _powerlaw_integral(jnp.hstack([egrid_[idx], Eb]), alpha1)
        pb2 = _powerlaw_integral(jnp.hstack([Eb, egrid_[idx + 1]]), alpha2)
        f.at[idx].set(pb1[0] + pb2[0])
        return K * f


class DoubleBrokenPL(AnaIntAdditive):
    pass


class DoubleSmoothlyBrokenPL(NumIntAdditive):
    pass


class CutoffPL(NumIntAdditive):
    r"""Power law with high-energy exponential cutoff.

    .. math::
        N(E) = K \left(\frac{E}{E_0}\right)^{-\alpha}
                \exp \left(-\frac{E}{E_\mathrm{c}}\right),

    where :math:`E_0` is the reference energy fixed at 1 keV.

    Parameters
    ----------
    alpha : Parameter, optional
        The power law photon index :math:`\alpha`, dimensionless.
    Ec : Parameter, optional
        The e-folding energy of exponential cutoff :math:`E_\mathrm{c}`,
        in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('alpha', r'\alpha', '', 1.0, -3.0, 10.0),
        ParamConfig('Ec', r'E_\mathrm{c}', 'keV', 15.0, 0.01, 1e4),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        alpha = params['alpha']
        Ec = params['Ec']
        K = params['K']
        return K * jnp.power(egrid, -alpha) * jnp.exp(-egrid / Ec)


class Compt(NumIntAdditive):
    r"""Power law with high-energy exponential cutoff, parametrized by the peak
    of :math:`\nu F_\nu`.

    .. math::
        N(E) = K \left(\frac{E}{E_0}\right)^{-\alpha}
                \exp \left[-\frac{(2-\alpha)E}{E_\mathrm{p}}\right],

    where :math:`E_0` is the reference energy fixed at 1 keV.

    Parameters
    ----------
    alpha : Parameter, optional
        The power law photon index :math:`\alpha`, dimensionless.
    Ep : Parameter, optional
        The peak energy :math:`E_\mathrm{p}` of :math:`\nu F_\nu`,
        in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('alpha', r'\alpha', '', 1.0, -3.0, 10.0),
        ParamConfig('Ep', r'E_\mathrm{p}', 'keV', 15.0, 0.01, 1e4),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        alpha = params['alpha']
        Ep = params['Ep']
        K = params['K']
        neg_inv_Ec = (alpha - 2.0) / Ep
        return K * jnp.power(egrid, -alpha) * jnp.exp(egrid * neg_inv_Ec)


class Gauss(NumIntAdditive):
    r"""Gaussian line profile.

    .. math::
        N(E) = \frac{K}{\sqrt{2\pi} \sigma}
                \exp\left[
                    -\frac{\left(E - E_\mathrm{l}\right)^2}{2 \sigma^2}
                \right].

    Parameters
    ----------
    El : Parameter, optional
        The line energy :math:`E_\mathrm{l}`, in units of keV.
    sigma : Parameter, optional
        The line width :math:`\sigma`, in units of keV.
    K : Parameter, optional
        The total photon flux :math:`K` of the line, in units of ph cm⁻² s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('El', r'E_\mathrm{l}', 'keV', 6.5, 0.0, 1e6),
        ParamConfig('sigma', r'\sigma', 'keV', 0.1, 0.0, 20),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        El = params['El']
        sigma = params['sigma']
        K = params['K']
        return K * stats.norm.pdf(x=egrid, loc=El, scale=sigma)


class LogParabola(NumIntAdditive):
    pass


class Lorentz(AnaIntAdditive):
    r"""Lorentzian line profile.

    .. math::
        N(E) = \frac{K/\mathcal{F}}{(E - E_\mathrm{l})^2 + (\sigma/2)^2},

    where

    .. math::
        \mathcal{F}=\int_0^{+\infty}
        \frac{1}{(E - E_\mathrm{l})^2 + (\sigma/2)^2} \ \mathrm{d}E.

    Parameters
    ----------
    El : Parameter, optional
        The line energy :math:`E_\mathrm{l}`, in units of keV.
    sigma : Parameter, optional
        The FWHM line width :math:`\sigma`, in units of keV.
    K : Parameter, optional
        The total photon flux :math:`K` of the line, in units of ph cm⁻² s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _config = (
        ParamConfig('El', r'E_\mathrm{l}', 'keV', 6.5, 1e-6, 1e6),
        ParamConfig('sigma', r'\sigma', 'keV', 0.1, 1e-3, 20),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        El = params['El']
        sigma = params['sigma']
        K = params['K']
        x = 2.0 * (egrid - El) / sigma
        integral = jnp.arctan(x)
        norm = 2.0 * K / (jnp.pi + 2 * jnp.arctan(2 * El / sigma))
        return norm * (integral[1:] - integral[:-1])


class OTTB(NumIntAdditive):
    r"""Optically-thin thermal bremsstrahlung.

    .. math::
        N(E) = K \left(\frac{E}{E_0}\right)^{-1} \exp\left(-\frac{E}{kT}\right)
                \exp\left(\frac{E_0}{kT}\right),

    where :math:`E_0` is the reference energy fixed at 1 keV.

    Parameters
    ----------
    kT : Parameter, optional
        The electron energy :math:`kT`, in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('kT', 'kT', 'keV', 30.0, 0.1, 1e3),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        kT = params['kT']
        K = params['K']
        return K * jnp.exp((1.0 - egrid) / kT) / egrid


class OTTS(NumIntAdditive):
    r"""Optically-thin thermal synchrotron [1]_.

    .. math::
        N(E) = K \exp\left[-\left(\frac{E}{E_\mathrm{c}}\right)^{1/3}\right].

    Parameters
    ----------
    Ec : Parameter, optional
        The energy scale :math:`E_\mathrm{c}`, in units of keV.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.

    References
    ----------
    .. [1] `Liang, E. P., et al., 1983, ApJ, 271, 776
           <https://adsabs.harvard.edu/full/1983ApJ...271..766L>`__
    """

    _config = (
        ParamConfig('Ec', r'E_\mathrm{c}', 'keV', 100.0, 1e-3, 1e3),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        Ec = params['Ec']
        K = params['K']
        return K * jnp.exp(-jnp.power(egrid / Ec, 1.0 / 3.0))


class PowerLaw(AnaIntAdditive):
    r"""Power law function.

    .. math::
        N(E) = K \left(\frac{E}{E_0}\right)^{-\alpha},

    where :math:`E_0` is the reference energy fixed at 1 keV.

    Parameters
    ----------
    alpha : Parameter, optional
        Power law photon index :math:`\alpha`, dimensionless.
    K : Parameter, optional
        Amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _config = (
        ParamConfig('alpha', r'\alpha', '', 1.01, -3.0, 10.0),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        return params['K'] * _powerlaw_integral(egrid, params['alpha'])


class PLFluxNorm(AnaIntAdditive):
    _args = ('emin', 'emax')
    _energy: bool

    def __init__(
        self,
        emin: float | int,
        emax: float | int,
        params: dict,
        latex: str | None,
    ):
        emin = float(emin)
        emax = float(emax)

        if emin >= emax:
            raise ValueError('emin must be less than emax')

        self._emin = emin
        self._emax = emax

        super().__init__(params, latex)

    @property
    def eval(self) -> CompEval:
        if self._integral_jit is None:
            emin = self._emin
            emax = self._emax
            energy = self._energy
            fn = jax.jit(self.integral)

            def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
                if energy:
                    keV_to_erg = 1.602176634e-9
                    params_ = {'alpha': params['alpha'] - 1.0}
                    f = fn(jnp.array([emin, emax]), params_)
                    factor = params['F'] / (f * keV_to_erg)
                else:
                    f = fn(jnp.array([emin, emax]), params)
                    factor = params['F'] / f

                return factor * fn(egrid, params)

            self._integral_jit = jax.jit(integral)

        return self._integral_jit

    @staticmethod
    def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        return _powerlaw_integral(egrid, params['alpha'])


class PLPhFlux(PLFluxNorm):
    r"""Power law function with photon flux used as normalization.

    .. math::
        N(E) &=
            \mathcal{F}_\mathrm{ph}
            \left[
                \int_{E_\mathrm{min}}^{E_\mathrm{max}}
                \left(\frac{E}{E_0}\right)^{-\alpha} \, \mathrm{d}E
            \right]^{-1}
            \left(\frac{E}{E_0}\right)^{-\alpha}\\
            &=
            \mathcal{F}_\mathrm{ph} (1 - \alpha)
            \left[
                \left(\frac{E_\mathrm{max}}{E_0}\right)^{1 - \alpha}
                - \left(\frac{E_\mathrm{min}}{E_0}\right)^{1 - \alpha}
            \right]^{-1}
            \left(\frac{E}{E_0}\right)^{-\alpha},

    where :math:`E_0` is the reference energy fixed at 1 keV.

    Parameters
    ----------
    emin : float or int
        Minimum energy of the band to calculate the flux, in units of keV.
    emax : float or int
        Maximum energy of the band to calculate the flux, in units of keV.
    alpha : Parameter, optional
        Power law photon index :math:`\alpha`, dimensionless.
    F : Parameter, optional
        Photon flux :math:`\mathcal{F}_\mathrm{ph}`, in units of ph cm⁻² s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _energy: bool = False
    _config = (
        ParamConfig('alpha', r'\alpha', '', 1.01, -3.0, 10.0),
        ParamConfig(
            'F', r'\mathcal{F}_\mathrm{ph}', 'ph cm^-2 s^-1', 1.0, 0.01, 1e10
        ),
    )


class PLEnFlux(PLFluxNorm):
    r"""Power law function with energy flux used as normalization.

    .. math::
        N(E) &=
            \mathcal{F}_\mathrm{en}
            \left[
                \int_{E_\mathrm{min}}^{E_\mathrm{max}}
                \left(\frac{E}{E_0}\right)^{-\alpha} \, E \, \mathrm{d}E
            \right]^{-1}
            \left(\frac{E}{E_0}\right)^{-\alpha}\\
            &=
            \mathcal{F}_\mathrm{en} (2 - \alpha)
            \left[
                \left(\frac{E_\mathrm{max}}{E_0}\right)^{2 - \alpha}
                - \left(\frac{E_\mathrm{min}}{E_0}\right)^{2 - \alpha}
            \right]^{-1}
            \left(\frac{E}{E_0}\right)^{-\alpha},

    where :math:`E_0` is the reference energy fixed at 1 keV.

    Parameters
    ----------
    emin : float or int
        Minimum energy of the band to calculate the flux, in units of keV.
    emax : float or int
        Maximum energy of the band to calculate the flux, in units of keV.
    alpha : Parameter, optional
        Power law photon index :math:`\alpha`, dimensionless.
    F : Parameter, optional
        Energy flux :math:`\mathcal{F}_\mathrm{en}`, in units of erg cm⁻² s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _energy: bool = True
    _config = (
        ParamConfig('alpha', r'\alpha', '', 1.01, -3.0, 10.0),
        ParamConfig(
            'F',
            r'\mathcal{F}_\mathrm{en}',
            'erg cm^-2 s^-1',
            1e-12,
            1e-30,
            1e30,
            log=True,
        ),
    )


class SmoothlyBrokenPL(NumIntAdditive):
    r"""Smoothly broken power law.

    .. math::
        N(E) = K
            \left( \frac{E}{E_0} \right) ^ {-\alpha_1}
            \left\{
                2\left[
                    1 + \left(\frac{E}{E_\mathrm{b}}
                        \right)^{\left( \alpha_2 - \alpha_1 \right) / \rho}
                \right]
            \right\}^{-\rho},

    where :math:`E_0` is the reference energy fixed at 1 keV.

    Parameters
    ----------
    alpha1 : Parameter, optional
        The low-energy power law index :math:`\alpha_1`, dimensionless.
    Eb : Parameter, optional
        The break energy :math:`E_\mathrm{b}`, in units of keV.
    alpha2 : Parameter, optional
        The high-energy power law index :math:`\alpha_2`, dimensionless.
    rho : Parameter, optional
        The smoothness parameter :math:`\rho`, dimensionless.
    K : Parameter, optional
        The amplitude :math:`K`, in units of ph cm⁻² s⁻¹ keV⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    method : {'trapz', 'simpson'}, optional
        Numerical integration method. Defaults to 'trapz'.
    """

    _config = (
        ParamConfig('alpha1', 'alpha_1', '', 1.0, -3.0, 10.0),
        ParamConfig('Eb', r'E_\mathrm{b}', 'keV', 5.0, 1e-2, 1e6),
        ParamConfig('alpha2', 'alpha_2', '', 2.0, -3.0, 10.0),
        ParamConfig('rho', r'\rho', '', 1.0, 1e-2, 1e2, fixed=True),
        ParamConfig('K', 'K', 'ph cm^-2 s^-1 keV^-1', 1.0, 1e-10, 1e10),
    )

    @staticmethod
    def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
        alpha1 = params['alpha1']
        Eb = params['Eb']
        alpha2 = params['alpha2']
        rho = params['rho']
        K = params['K']

        e = egrid / Eb
        x = (alpha2 - alpha1) / rho * jnp.log(e)

        threshold = 30
        alpha = jnp.where(x > threshold, alpha2, alpha1)
        r = jnp.where(jnp.abs(x) > threshold, 0.5, 0.5 * (1.0 + jnp.exp(x)))

        return K * jnp.power(e, -alpha) * jnp.power(r, -rho)



"""Convolution models."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp

from elisa.models.model import ConvolutionComponent, ParamConfig

if TYPE_CHECKING:
    from collections.abc import Callable

    from elisa.util.typing import ConvolveEval, JAXArray, NameValMapping

__all__ = ['EnFlux', 'PhFlux', 'ZAShift', 'ZMShift', 'VAShift', 'VMShift']


class NormConvolution(ConvolutionComponent):
    _args = ('emin', 'emax')
    _kwargs = ('ngrid', 'elog')
    _supported = frozenset({'add'})

    def __init__(
        self,
        emin: float | int,
        emax: float | int,
        params: dict,
        latex: str | None,
        ngrid: int | None,
        elog: bool | None,
    ):
        self._emin = float(emin)
        self.emax = emax

        self.ngrid = 1000 if ngrid is None else ngrid
        self.elog = True if elog is None else bool(elog)

        self._prev_config: tuple | None = None

        super().__init__(params, latex)

    @staticmethod
    @abstractmethod
    def convolve(
        egrid: JAXArray,
        params: NameValMapping,
        model_fn: Callable[[JAXArray], JAXArray],
        flux_egrid: JAXArray,
    ) -> JAXArray:
        """Convolve a model function.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : dict
            Parameter dict for the convolution model.
        model_fn : callable
            The model function to be convolved, which takes the energy grid as
            input and returns the model flux over the grid.
        flux_egrid : ndarray
            Photon energy grid used to calculate flux in units of keV.

        Returns
        -------
        value : ndarray
            The re-normalized model over `egrid`, in units of
            ph cm⁻² s⁻¹ keV⁻¹.
        """
        pass

    @property
    def eval(self) -> ConvolveEval:
        if self._prev_config == (self.emin, self.emax, self.ngrid, self.elog):
            return self._convolve_jit

        else:
            if self.elog:
                flux_egrid = jnp.geomspace(self.emin, self.emax, self.ngrid)
            else:
                flux_egrid = jnp.linspace(self.emin, self.emax, self.ngrid)

            fn = self.convolve

            def convolve(
                egrid: JAXArray,
                params: NameValMapping,
                model_fn: Callable[[JAXArray], JAXArray],
            ) -> JAXArray:
                # TODO: egrid can be reused to reduce computation
                return fn(egrid, params, model_fn, flux_egrid)

            self._prev_config = (self.emin, self.emax, self.ngrid, self.elog)
            self._convolve_jit = jax.jit(convolve, static_argnums=2)

            return self._convolve_jit

    @property
    def emin(self) -> float:
        """Minimum value of photon energy grid"""
        return self._emin

    @emin.setter
    def emin(self, value: float | int):
        value = float(value)
        if value >= self.emax:
            raise ValueError('emin must be less than emax')
        self._emin = value

    @property
    def emax(self) -> float:
        """Maximum value of photon energy grid"""
        return self._emax

    @emax.setter
    def emax(self, value: float | int):
        value = float(value)
        if value <= self._emin:
            raise ValueError('emax must be larger than emin')
        self._emax = float(value)

    @property
    def ngrid(self) -> int:
        """Photon energy grid number."""
        return self._ngrid

    @ngrid.setter
    def ngrid(self, value: int):
        self._ngrid = int(value)

    @property
    def elog(self) -> bool:
        """Whether to use logarithmically regular energy grids."""
        return self._elog

    @elog.setter
    def elog(self, value: bool):
        self._elog = bool(value)


class PhFlux(NormConvolution):
    r"""Normalize an additive model by photon flux between `emin` and `emax`.

    .. math::
        N'(E) =
            \mathcal{F}_\mathrm{ph}
            \left[\int_{E_\mathrm{min}}^{E_\mathrm{max}} N(E) \, dE\right]^{-1}
            N(E)

    .. warning::
        The normalization of one of the additive components **must** be fixed
        to a positive value.

    .. warning::
        The flux is calculated by trapezoidal rule, and is accurate only if
        enough numbers of energy grids are used.

    Parameters
    ----------
    emin : float or int
        Minimum energy of the band to calculate the flux, in units of keV.
    emax : float or int
        Maximum energy of the band to calculate the flux, in units of keV.
    F : Parameter, optional
        Photon flux :math:`\mathcal{F}_\mathrm{ph}`, in units of ph cm⁻² s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    ngrid : int, optional
        The energy grid number to use. The default is 1000.
    elog : bool, optional
        Whether to use logarithmically regular energy grids.
        The default is True.
    """

    _config = (
        ParamConfig(
            'F', r'\mathcal{F}_\mathrm{ph}', 'ph cm^-2 s^-1', 1.0, 0.01, 1e10
        ),
    )

    @staticmethod
    def convolve(
        egrid: JAXArray,
        params: NameValMapping,
        model_fn: Callable[[JAXArray], JAXArray],
        flux_egrid: JAXArray,
    ) -> JAXArray:
        F = params['F']
        mflux = jnp.sum(model_fn(flux_egrid))
        flux = model_fn(egrid)
        return F / mflux * flux


class EnFlux(NormConvolution):
    r"""Normalize an additive model by energy flux between `emin` and `emax`.

    .. math::
        N'(E) =
            \mathcal{F}_\mathrm{en}
            \left[\int_{E_\mathrm{min}}^{E_\mathrm{max}} EN(E)\,dE\right]^{-1}
            N(E)

    .. warning::
        The normalization of one of the additive components **must** be fixed
        to a positive value.

    .. warning::
        The flux is calculated by trapezoidal rule, and is accurate only if
        enough numbers of energy grids are used.

    Parameters
    ----------
    emin : float or int
        Minimum energy of the band to calculate the flux, in units of keV.
    emax : float or int
        Maximum energy of the band to calculate the flux, in units of keV.
    F : Parameter, optional
        Energy flux :math:`\mathcal{F}_\mathrm{en}`, in units of erg cm⁻² s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    ngrid : int, optional
        The energy grid number to use. The default is 1000.
    elog : bool, optional
        Whether to use logarithmically regular energy grids.
        The default is True.
    """

    _config = (
        ParamConfig(
            'F',
            r'\mathcal{F}_\mathrm{en}',
            'erg cm^-2 s^-1',
            1e-12,
            1e-30,
            1e30,
            log=True,
        ),
    )

    @staticmethod
    def convolve(
        egrid: JAXArray,
        params: NameValMapping,
        model_fn: Callable[[JAXArray], JAXArray],
        flux_egrid: JAXArray,
    ) -> JAXArray:
        F = params['F']
        keV_to_erg = 1.602176634e-9
        mid = jnp.sqrt(flux_egrid[:-1] * flux_egrid[1:])
        _flux = model_fn(flux_egrid)
        mflux = jnp.sum(keV_to_erg * mid * _flux)
        flux = model_fn(egrid)
        return F / mflux * flux


class ZAShift(ConvolutionComponent):
    r"""Redshifts an additive model.

    Consider a source with an emission area of radius :math:`R` at redshift
    :math:`z`. Given the flux function :math:`N(E)` [ph s⁻¹ cm⁻² keV⁻¹] at the
    radius :math:`R`, the observed number of photons :math:`n` between the
    energy range :math:`e_1` [keV] and :math:`e_2` [keV] during an exposure
    time of :math:`\Delta t` [s] is calculated as follows:

    .. math::
        n &= \frac{R^2}{{D_\mathrm{c}}^2} \frac{\Delta t}{1+z}
             \int_{e_1(1+z)}^{e_2(1+z)} N(E) \, \mathrm{d}E
          \\\\
          &= \frac{R^2}{{D_\mathrm{c}}^2} \frac{\Delta t}{1+z}
             \int_{E_1}^{E_2} N(E) \, \mathrm{d}E,

    where :math:`E_1 = e_1 (1+z)` [keV], :math:`E_2 = e_2 (1+z)` [keV] and
    :math:`D_\mathrm{c}` is the comoving distance of the source at redshift
    :math:`z`.

    Note that the :math:`\frac{R^2}{{D_\mathrm{c}}^2}` factor is absorbed into
    the normalization of :math:`N(E)` in practice.

    Parameters
    ----------
    z : Parameter, optional
        Redshift :math:`z`, dimensionless.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _supported = frozenset({'add'})
    _config = (ParamConfig('z', 'z', '', 0.0, -0.999, 15.0, fixed=True),)

    @staticmethod
    def convolve(
        egrid: JAXArray,
        params: NameValMapping,
        model_fn: Callable[[JAXArray], JAXArray],
    ) -> JAXArray:
        factor = 1.0 + params['z']
        return model_fn(egrid * factor) / factor


class ZMShift(ConvolutionComponent):
    r"""Redshifts a multiplicative model.

    Consider a source at redshift :math:`z`. Given the dimensionless model
    function :math:`M(E)`, the observed value between the energy range
    :math:`e_1` [keV] and :math:`e_2` [keV] is calculated as follows:

    .. math::
        m &= \frac{1}{(e_2 - e_1)(1+z)}
             \int_{e_1(1+z)}^{e_2(1+z)} M(E) \, \mathrm{d}E
          \\\\
          &= \frac{1}{E_2 - E_1} \int_{E_1}^{E_2} M(E) \, \mathrm{d}E,

    where :math:`E_1 = e_1 (1+z)` [keV] and :math:`E_2 = e_2 (1+z)` [keV].

    Parameters
    ----------
    z : Parameter, optional
        Redshift :math:`z`, dimensionless.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _supported = frozenset({'mul'})
    _config = (ParamConfig('z', 'z', '', 0.0, -0.999, 15.0, fixed=True),)

    @staticmethod
    def convolve(
        egrid: JAXArray,
        params: NameValMapping,
        model_fn: Callable[[JAXArray], JAXArray],
    ) -> JAXArray:
        factor = 1.0 + params['z']
        return model_fn(egrid * factor)


class VAShift(ConvolutionComponent):
    r"""Velocity shifts an additive model.

    Consider a source with an emission area of radius :math:`R`, moving with
    speed :math:`v` along line of sight. Given the flux function :math:`N(E)`
    [ph s⁻¹ cm⁻² keV⁻¹] at the radius :math:`R`, the observed number of photons
    :math:`n` between the energy range :math:`e_1` [keV] and :math:`e_2` [keV]
    during an exposure time of :math:`\Delta t` [s] is calculated as follows:

    .. math::
        n &= \Delta t \int_{fe_1}^{fe_2} N(E) \, \mathrm{d}E
          \\\\
          &= \Delta t \int_{E_1}^{E_2} N(E) \, \mathrm{d}E,

    where :math:`E_1 = f e_1` [keV], :math:`E_2 = f e_2` [keV], and
    :math:`f = 1 - v/c`.

    Parameters
    ----------
    v : Parameter, optional
        Velocity :math:`v`, in units of km s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _supported = frozenset({'add'})
    _config = (ParamConfig('v', 'v', 'km s^-1', 0.0, -1e4, 1e4, fixed=True),)

    @staticmethod
    def convolve(
        egrid: JAXArray,
        params: NameValMapping,
        model_fn: Callable[[JAXArray], JAXArray],
    ) -> JAXArray:
        v = params['v']  # unit: km/s
        c = 299792.458  # unit: km/s
        f = 1.0 - v / c
        return model_fn(egrid * f)


class VMShift(ConvolutionComponent):
    r"""Velocity shifts a multiplicative model.

    Consider a source moving with speed :math:`v` along line of sight. Given
    the dimensionless model function :math:`M(E)`, the observed value between
    the energy range :math:`e_1` [keV] and :math:`e_2` [keV] is calculated as
    follows:

    .. math::
        m &= \frac{1}{f (e_2 - e_1)} \int_{f e_1}^{f e_2} M(E) \, \mathrm{d}E
          \\\\
          &= \frac{1}{E_2 - E_1} \int_{E_1}^{E_2} M(E) \, \mathrm{d}E,

    where :math:`E_1 = f e_1` [keV], :math:`E_2 = f e_2` [keV], and
    :math:`f = 1 - v/c`.

    Parameters
    ----------
    v : Parameter, optional
        Velocity :math:`v`, in units of km s⁻¹.
    latex : str, optional
        :math:`\LaTeX` format of the component. Defaults to class name.
    """

    _supported = frozenset({'mul'})
    _config = (ParamConfig('v', 'v', 'km s^-1', 0.0, -1e4, 1e4, fixed=True),)

    @staticmethod
    def convolve(
        egrid: JAXArray,
        params: NameValMapping,
        model_fn: Callable[[JAXArray], JAXArray],
    ) -> JAXArray:
        v = params['v']  # unit: km/s
        c = 299792.458  # unit: km/s
        f = 1.0 - v / c
        return model_fn(egrid * f)



"""Visualize fit and analysis results."""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import arviz as az
import jax
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stats

from elisa.infer.helper import check_params
from elisa.plot.data import MLEPlotData, PosteriorPlotData
from elisa.plot.misc import plot_corner, plot_trace
from elisa.plot.scale import LinLogScale, get_scale
from elisa.plot.util import get_colors, get_markers

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any, Literal

    from matplotlib.pyplot import Axes, Figure

    from elisa.infer.results import FitResult, MLEResult, PosteriorResult
    from elisa.plot.data import PlotData
    from elisa.util.typing import Array, NumPyArray


def _plot_step(
    ax: Axes, x_left: Array, x_right: Array, y: Array, **step_kwargs
) -> None:
    assert len(y) == len(x_left) == len(x_right)

    step_kwargs['where'] = 'post'

    mask = x_left[1:] != x_right[:-1]
    idx = np.insert(np.flatnonzero(mask) + 1, 0, 0)
    idx = np.append(idx, len(y))
    for i in range(len(idx) - 1):
        i_slice = slice(idx[i], idx[i + 1])
        x_slice = np.append(x_left[i_slice], x_right[i_slice][-1])
        y_slice = y[i_slice]
        y_slice = np.append(y_slice, y_slice[-1])
        ax.step(x_slice, y_slice, **step_kwargs)


def _plot_ribbon(
    ax,
    x_left: Array,
    x_right: Array,
    y_ribbons: Sequence[Array],
    **ribbon_kwargs,
) -> None:
    y_ribbons = list(map(np.asarray, y_ribbons))
    shape = y_ribbons[0].shape
    assert len(shape) == 2 and shape[0] == 2
    assert shape[1] == len(x_left) == len(x_right)
    assert all(ribbon.shape == shape for ribbon in y_ribbons)

    ribbon_kwargs['step'] = 'post'

    mask = x_left[1:] != x_right[:-1]
    idx = np.insert(np.flatnonzero(mask) + 1, 0, 0)
    idx = np.append(idx, shape[1])
    for i in range(len(idx) - 1):
        i_slice = slice(idx[i], idx[i + 1])
        x_slice = np.append(x_left[i_slice], x_right[i_slice][-1])

        for ribbon in y_ribbons:
            lower = ribbon[0]
            lower_slice = lower[i_slice]
            lower_slice = np.append(lower_slice, lower_slice[-1])
            upper = ribbon[1]
            upper_slice = upper[i_slice]
            upper_slice = np.append(upper_slice, upper_slice[-1])
            ax.fill_between(x_slice, lower_slice, upper_slice, **ribbon_kwargs)


def _adjust_log_range(
    ax: Axes,
    axis: Literal['x', 'y'] = 'y',
    octave: int = 4,
) -> None:
    octave = round(octave)
    assert octave > 0
    if axis == 'y':
        vmin, vmax = ax.dataLim.intervaly
        set_lim = ax.set_ylim
    else:
        vmin, vmax = ax.dataLim.intervalx
        set_lim = ax.set_xlim

    if np.log10(vmax / max(1e-30, vmin)) > int(octave):
        vmin = vmax / 10**octave

    set_lim(
        np.power(10, np.log10(vmin) - 0.05), np.power(10, np.log10(vmax) + 0.1)
    )


def _get_qq(
    q: NumPyArray,
    detrend: bool,
    cl: float,
    qsim: NumPyArray | None = None,
) -> tuple[NumPyArray, ...]:
    """Get the Q-Q and pointwise confidence/credible interval.

    References
    ----------
    .. [1] doi:10.1080/00031305.2013.847865
    """
    # https://stats.stackexchange.com/a/9007
    # https://stats.stackexchange.com/a/152834
    alpha = np.pi / 8  # 3/8 is also ok
    n = len(q)
    theor = stats.norm.ppf((np.arange(1, n + 1) - alpha) / (n - 2 * alpha + 1))

    q = np.sort(q)
    if qsim is not None:
        line, lower, upper = np.quantile(
            np.sort(qsim, axis=1),
            q=[0.5, 0.5 - 0.5 * cl, 0.5 + 0.5 * cl],
            axis=0,
        )
    else:
        line = np.array(theor)
        grid = np.arange(1, n + 1)
        lower = stats.beta.ppf(0.5 - cl * 0.5, grid, n + 1 - grid)
        upper = stats.beta.ppf(0.5 + cl * 0.5, grid, n + 1 - grid)
        lower = stats.norm.ppf(lower)
        upper = stats.norm.ppf(upper)

    if detrend:
        q -= theor
        line -= theor
        lower -= theor
        upper -= theor

    return theor, q, line, lower, upper


def _get_pit_ecdf(
    pit: NumPyArray,
    cl: float,
    detrend: bool,
) -> tuple[NumPyArray, ...]:
    """Get the empirical CDF of PIT and pointwise confidence/credible interval.

    References
    ----------
    .. [1] doi:10.1007/s11222-022-10090-6
    """
    n = len(pit)

    # See ref [1] for the following
    scaled_rank = np.linspace(0.0, 1.0, n + 1)
    pit_ecdf = np.count_nonzero(pit <= scaled_rank[:, None], axis=1) / n
    lower, upper = stats.binom.interval(cl, n, scaled_rank)
    lower = lower / n
    upper = upper / n

    if detrend:
        line = np.zeros_like(scaled_rank)
        lower -= scaled_rank
        upper -= scaled_rank
        pit_ecdf -= scaled_rank
    else:
        line = scaled_rank

    return scaled_rank, pit_ecdf, line, lower, upper

    # x = np.hstack([0.0, np.sort(pit), 1.0])
    # pit_ecdf = np.hstack([0.0, np.arange(n) / n, 1.0])
    # lower, upper = stats.binom.interval(cl, n, x)
    # lower = lower / n
    # upper = upper / n
    #
    # if detrend:
    #     pit_ecdf -= x
    #     lower -= x
    #     upper -= x
    #     line = np.zeros_like(x)
    # else:
    #     line = x
    #
    # return x, pit_ecdf, line, lower, upper


# def _get_pit_pdf(pit_intervals: NumPyArray) -> NumPyArray:
#     """Get the pdf of PIT.
#
#     References
#     ----------
#     .. [1] doi:10.1111/j.1541-0420.2009.01191.x
#     """
#     assert len(pit_intervals.shape) == 2 and pit_intervals.shape[1] == 2
#
#     grid = np.unique(pit_intervals)
#     if grid[0] > 0.0:
#         grid = np.insert(grid, 0, 0)
#     if grid[-1] < 1.0:
#         grid = np.append(grid, 1.0)
#
#     n = len(pit_intervals)
#     mask = pit_intervals[:, 0] != pit_intervals[:, 1]
#     cover_mask = np.bitwise_and(
#         pit_intervals[:, :1] <= grid[:-1],
#         grid[1:] <= pit_intervals[:, 1:],
#     )
#     pdf = np.zeros((n, len(grid) - 1))
#     pdf[cover_mask] = np.repeat(
#         1.0 / (pit_intervals[mask, 1] - pit_intervals[mask, 0]),
#         np.count_nonzero(cover_mask[mask], axis=1),
#     )
#     idx = np.clip(grid.searchsorted(pit_intervals[~mask, 0]) - 1, 0, None)
#     pdf[~mask, idx] = 1.0 / (grid[idx + 1] - grid[idx])
#     return pdf.mean(0)


class PlotConfig:
    """Plotting configuration."""

    _YLABLES = {
        'ce': r'$C_E\ \mathrm{[counts\ s^{-1}\ keV^{-1}]}$',
        'ne': r'$N_E\ \mathrm{[ph\ cm^{-2}\ s^{-1}\ keV^{-1}]}$',
        'ene': r'$E N_E\ \mathrm{[erg\ cm^{-2}\ s^{-1}\ keV^{-1}]}$',
        'Fv': r'$F_{\nu}\ \mathrm{[erg\ cm^{-2}\ s^{-1}\ keV^{-1}]}$',
        'eene': r'$E^2 N_E\ \mathrm{[erg\ cm^{-2}\ s^{-1}]}$',
        'vFv': r'$\nu F_{\nu}\ \mathrm{[erg\ cm^{-2}\ s^{-1}]}$',
        'rd': r'$r_D\ [\mathrm{\sigma}]$',
        'rp': r'$r_\mathrm{P}\ [\mathrm{\sigma}]$',
        'rq': r'$r_Q\ [\mathrm{\sigma}]$',
    }

    def __init__(
        self,
        alpha: float = 0.8,
        palette: Any = 'colorblind',
        xscale: Literal['linear', 'log'] = 'log',
        yscale: Literal['linear', 'log', 'linlog'] = 'linlog',
        lin_frac: float = 0.15,
        cl: tuple[float, ...] = (0.683, 0.95),
        residuals: Literal['rd', 'rp', 'rq'] = 'rq',
        random_quantile: bool = False,
        mark_outlier_residuals: bool = False,
        residuals_ci_with_sign: bool = True,
        fill_residuals_ci: bool = True,
        plot_comps: bool = False,
        seed: int | None = None,
    ):
        self.alpha = alpha
        self.palette = palette
        self.xscale = xscale
        self.yscale = yscale
        self.lin_frac = lin_frac
        self.cl = cl
        self.residuals = residuals
        self.random_quantile = random_quantile
        self.mark_outlier_residuals = mark_outlier_residuals
        self.residuals_ci_with_sign = residuals_ci_with_sign
        self.fill_residuals_ci = fill_residuals_ci
        self.plot_comps = plot_comps
        self.seed = seed

    @property
    def alpha(self) -> float:
        """Transparency of colors."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: float):
        alpha = float(alpha)
        if not 0.0 < alpha <= 1.0:
            raise ValueError('alpha must be in (0, 1]')
        self._alpha = alpha

    @property
    def palette(self) -> Any:
        """Color palettes, see [1]_ for details.

        References
        ----------
        .. [1] `seaborn tutorial: Choosing color palettes
                <https://seaborn.pydata.org/tutorial/color_palettes.html>`__
        """
        return self._palette

    @palette.setter
    def palette(self, palette: Any):
        self._palette = palette

    @property
    def xscale(self) -> Literal['linear', 'log']:
        """X-axis scale of spectral plot.

        Should be ``'linear'``, or ``'log'``.
        """
        return self._xscale

    @xscale.setter
    def xscale(self, xscale: Literal['linear', 'log']):
        if xscale not in {'linear', 'log'}:
            raise ValueError('xscale must be "linear" or "log"')
        self._xscale = xscale

    @property
    def yscale(self) -> Literal['linear', 'log', 'linlog']:
        """X-axis scale of spectral plot.

        Should be ``'linear'``, ``'log'``, or ``'linlog'``.
        """
        return self._yscale

    @yscale.setter
    def yscale(self, yscale: Literal['linear', 'log', 'linlog']):
        if yscale not in {'linear', 'log', 'linlog'}:
            raise ValueError('yscale must be "linear", "log", or "linlog"')
        self._yscale = yscale

    @property
    def lin_frac(self) -> float:
        """Linear fraction of the ``linlog`` plot."""
        return self._lin_frac

    @lin_frac.setter
    def lin_frac(self, lin_frac: float):
        lin_frac = float(lin_frac)
        if not 0.0 < lin_frac <= 0.5:
            raise ValueError('lin_frac must be in (0, 0.5]')
        self._lin_frac = lin_frac

    @property
    def cl(self) -> NumPyArray:
        """Confidence/Credible level."""
        return self._cl

    @cl.setter
    def cl(self, cl: float | Sequence[float]):
        cl = np.sort(np.atleast_1d(cl)).astype(float)
        for c in cl:
            if not 0.0 < c < 1.0:
                raise ValueError('cl must be in (0, 1)')
        self._cl = cl

    @property
    def residuals(self) -> Literal['rd', 'rp', 'rq']:
        """Default type of residual plot."""
        return self._residuals

    @residuals.setter
    def residuals(self, residuals: Literal['rd', 'rp', 'rq']):
        if residuals not in {'rd', 'rp', 'rq'}:
            raise ValueError(
                'residuals type must be "rd" (deviance), "rp" (pearson), or '
                '"rq" (quantile)'
            )
        self._residuals = residuals

    @property
    def random_quantile(self) -> bool:
        """Whether to randomize the quantile residual."""
        return self._random_quantile

    @random_quantile.setter
    def random_quantile(self, random_quantile: bool):
        self._random_quantile = bool(random_quantile)

    @property
    def mark_outlier_residuals(self) -> bool:
        """Whether to mark outlier residuals with red crosses."""
        return self._mark_outlier_residuals

    @mark_outlier_residuals.setter
    def mark_outlier_residuals(self, mark_outlier_residuals: bool):
        self._mark_outlier_residuals = bool(mark_outlier_residuals)

    @property
    def residuals_ci_with_sign(self) -> bool:
        """Whether to take account residuals' sign when calculate CI bands."""
        return self._residuals_ci_with_sign

    @residuals_ci_with_sign.setter
    def residuals_ci_with_sign(self, residuals_ci_with_sign: bool):
        self._residuals_ci_with_sign = bool(residuals_ci_with_sign)

    @property
    def fill_residuals_ci(self) -> bool:
        """Whether to fill residuals' CI bands."""
        return self._fill_residuals_ci

    @fill_residuals_ci.setter
    def fill_residuals_ci(self, fill_residuals_ci: bool):
        self._fill_residuals_ci = bool(fill_residuals_ci)

    @property
    def plot_comps(self) -> bool:
        """Whether to plot additive components in spectral plot."""
        return self._plot_comps

    @plot_comps.setter
    def plot_comps(self, plot_comps: bool):
        self._plot_comps = bool(plot_comps)

    @property
    def seed(self) -> int | None:
        """Random seed used in calculation."""
        return self._seed

    @seed.setter
    def seed(self, seed: int | None):
        if seed is not None:
            seed = int(seed)
        self._seed = seed


class Plotter(ABC):
    """Plotter to visualize fit results."""

    _colors: dict | None = None
    _palette: Any | None = None
    _comps_latex: dict[str, str] | None = None
    _params_latex: dict[str, str] | None = None
    _supported: tuple[str, ...]
    data: dict[str, PlotData] | None = None

    def __init__(self, result: FitResult, config: PlotConfig = None):
        self._result = result
        self.data = self.get_plot_data(result)
        self.config = config
        markers = get_markers(len(self.data))
        self._markers = dict(zip(self.data.keys(), markers, strict=True))

    @abstractmethod
    def __call__(self, plots: str = 'data ne r') -> dict[str, Figure]:
        pass

    @abstractmethod
    def plot_corner(self, *args, **kwargs) -> Figure:
        """Corner plot of bootstrap/posterior parameters."""
        pass

    @staticmethod
    @abstractmethod
    def get_plot_data(result: FitResult) -> dict[str, PlotData]:
        """Get PlotData from FitResult."""
        pass

    @property
    def config(self) -> PlotConfig:
        """Plotting configuration."""
        return self._config

    @config.setter
    def config(self, config: PlotConfig):
        if config is None:
            config = PlotConfig()
        elif not isinstance(config, PlotConfig):
            raise TypeError('config must be a PlotConfig instance')

        self._config = config

    def set_colors(self, colors: dict[str, Any] | None = None):
        """Specify the colors of data points used in the plots.

        Parameters
        ----------
        colors : dict or None, optional
            If a dict is provided, will use the provided values as colors.
            If None, the default colors will be used. The default is None.
        """
        if colors is None:
            self._palette = None
            self._colors = None
        else:
            colors = dict(colors)
            if not set(self.data.keys()).issubset(colors.keys()):
                missing = ', '.join(i for i in self.data if i not in colors)
                raise ValueError(f'missing colors for those data: {missing}')
            self._palette = 'customized'
            self._colors = colors

    @property
    def colors(self):
        """Plotting color for each data."""
        if self._palette not in ('customized', self.config.palette):
            colors = get_colors(len(self.data), palette=self.config.palette)
            self._palette = self.config.palette
            self._colors = dict(zip(self.data.keys(), colors, strict=True))
        return self._colors

    @property
    def ndata(self):
        """Data points number."""
        ndata = {name: data.ndata for name, data in self.data.items()}
        ndata['total'] = sum(ndata.values())
        return ndata

    @property
    def comps_latex(self) -> dict[str, str]:
        """LaTeX representation of components."""
        if self._comps_latex is None:
            self._comps_latex = {
                k: f'${v}$ ' if v else ''
                for k, v in self._result._helper.params_comp_latex.items()
            }
        return self._comps_latex

    @property
    def params_latex(self) -> dict[str, str]:
        """LaTeX representation of parameters."""
        if self._params_latex is None:
            self._params_latex = {
                k: f'${v}$'
                for k, v in self._result._helper.params_latex.items()
            }
        return self._params_latex

    @property
    def params_unit(self) -> dict[str, str]:
        """Unit of parameters."""
        return self._result._helper.params_unit

    @property
    def params_titles(self) -> dict[str, str]:
        """Title of parameters."""
        comps_latex = self.comps_latex
        params_latex = self.params_latex
        params = self._result._helper.params_names['all']
        return {p: comps_latex[p] + params_latex[p] for p in params}

    @property
    def params_labels(self) -> dict[str, str]:
        """Label of parameters."""
        comps_latex = self.comps_latex
        params_latex = self.params_latex
        params_unit = {
            k: f'\n[{v}]' if v else v for k, v in self.params_unit.items()
        }
        params = self._result._helper.params_names['all']
        return {
            p: comps_latex[p] + params_latex[p] + params_unit[p]
            for p in params
        }

    def set_xlabel(self, ax: Axes):
        ax.set_xlabel(r'$\mathrm{Energy\ [keV]}$')

    def plot_spec(
        self,
        data: bool = True,
        ne: bool = True,
        ene: bool = False,
        eene: bool = False,
        residuals: bool | Literal['rd', 'rp', 'rq'] = True,
        *,
        egrid: Mapping[str, NumPyArray] | None = None,
        params: Mapping[str, float | int | Array] | None = None,
        label_Fv: bool = False,
        label_vFv: bool = False,
    ) -> Figure:
        r"""Spectral plot.

        Parameters
        ----------
        data : bool, optional
            Whether to plot folded model and data. The default is ``True``.
        ne : bool, optional
            Whether to plot :math:`N(E)`. The default is ``True``.
        ene : bool, optional
            Whether to plot :math:`E N(E)`. The default is ``False``.
        eene : bool, optional
            Whether to plot :math:`E^2 N(E)`. The default is ``False``.
        residuals : bool or {'rd', 'rp', 'rq'}, optional
            Whether to plot residuals. Available options are:

                * ``True``: plot default residuals
                * ``False``: do not plot residuals
                * ``'rd'``: plot deviance residuals
                * ``'rp'``: plot Pearson residuals
                * ``'rq'``: plot quantile residuals

            The default is ``True``.
        egrid : dict, optional
            Overwrite the photon energy grid when plotting unfolded model.
        params : dict, optional
            Overwrite the photon energy grid when plotting unfolded model.
        label_Fv : bool, optional
            Whether to label the y-axis of :math:`E N(E)` plot as
            :math:`F_{\nu}`. The default is ``False``.
        label_vFv : bool, optional
            Whether to label the y-axis of :math:`E^2 N(E)` plot as
            :math:`\nu F_{\nu}`. The default is ``False``.

        Returns
        -------
        Figure
            The Figure object containing the spectral plot.
        """
        nrows = data + ne + ene + eene + bool(residuals)
        height_ratios = [1.618] * nrows
        if residuals:
            height_ratios[-1] = 1.0

        config = self.config
        fig, axs = plt.subplots(
            nrows=nrows,
            ncols=1,
            sharex='all',
            height_ratios=height_ratios,
            gridspec_kw={'bottom': 0.07, 'top': 0.97, 'hspace': 0.03},
            figsize=(8, 4 + nrows),
            squeeze=False,
        )
        axs = axs.ravel()
        fig.align_ylabels(axs)

        for ax in axs:
            ax.tick_params(
                axis='both',
                which='both',
                direction='in',
                bottom=True,
                top=True,
                left=True,
                right=True,
            )

        plt.rcParams['axes.formatter.min_exponent'] = 3

        self.set_xlabel(axs[-1])

        plots = []
        if data:
            plots.append('ce')
        if ne:
            plots.append('ne')
        if ene:
            plots.append('ene')
        if eene:
            plots.append('eene')

        residuals: Literal['rd', 'rp', 'rq'] | None
        if residuals:
            plots.append('residuals')
            if residuals is True:
                residuals = config.residuals
        else:
            residuals = None

        axs_dict = dict(zip(plots, axs, strict=True))

        yscale = config.yscale

        if data:
            ax = axs_dict['ce']
            self.plot_ce(ax)
            self.plot_folded(ax)
            if yscale == 'linear':
                ax.set_yscale('linear')
            else:
                ax.set_yscale('log')
                dmin, dmax = ax.get_yaxis().get_data_interval()
                vmin = ax.get_ylim()[0]
                if yscale == 'linlog' and dmin <= 0.0:
                    lin_frac = config.lin_frac
                    if np.log10(dmax / vmin) > 7:
                        vmin = 1e-7 * dmax
                    scale = LinLogScale(
                        axis=None,
                        base=10.0,
                        lin_thresh=vmin,
                        lin_scale=get_scale(10.0, vmin, dmin, dmax, lin_frac),
                    )
                    ax.set_yscale(scale)
                    ax.axhline(vmin, c='k', lw=0.15, ls=':', zorder=-1)
                else:
                    _adjust_log_range(ax, 'y', 7)
        if ne:
            self.plot_unfolded(axs_dict['ne'], 'ne', params, egrid)
            if yscale != 'linear':
                axs_dict['ne'].set_yscale('log')
                _adjust_log_range(axs_dict['ne'], 'y')
        if ene:
            self.plot_unfolded(
                axs_dict['ene'], 'ene', params, egrid, label_Fv=label_Fv
            )
            if yscale != 'linear':
                axs_dict['ene'].set_yscale('log')
                _adjust_log_range(axs_dict['ene'], 'y')
        if eene:
            self.plot_unfolded(
                axs_dict['eene'], 'eene', params, egrid, label_vFv=label_vFv
            )
            if yscale != 'linear':
                axs_dict['eene'].set_yscale('log')
                _adjust_log_range(axs_dict['eene'], 'y')
        if residuals:
            self.plot_residuals(axs_dict['residuals'], residuals)

        axs[0].set_xscale(config.xscale)
        intervalx = np.array([ax.dataLim.intervalx for ax in axs])
        xmin = intervalx[:, 0].min()
        xmax = intervalx[:, 1].max()
        axs[0].set_xlim(xmin * 0.97, xmax * 1.06)

        for ax in axs:
            ax.relim()
            ax.autoscale_view()

        return fig

    def plot_unfolded(
        self,
        ax: Axes,
        mtype: Literal['ne', 'ene', 'eene'],
        params: Mapping[str, float | int | Array] | None = None,
        egrid: Mapping[str, NumPyArray] | None = None,
        label_Fv: bool = False,
        label_vFv: bool = False,
    ):
        r"""Plot unfolded model.

        Parameters
        ----------
        ax : Axes
            The Axes object to plot.
        mtype : {'ne', 'ene', 'eene'}
            The type of unfolded model, available options are:

                * ``'ne'``: plot :math:`N(E)`
                * ``'ene'``: plot :math:`E N(E)`
                * ``'eene'``: plot :math:`E^2 N(E)`

        params : dict, optional
            Overwrite the parameters when plotting unfolded model.
        egrid : dict, optional
            Overwrite the photon energy grid when plotting unfolded model.
        label_Fv : bool, optional
            Whether to label the y-axis of :math:`E N(E)` plot as
            :math:`F_{\nu}`. The default is ``False``.
        label_vFv : bool, optional
            Whether to label the y-axis of :math:`E^2 N(E)` plot as
            :math:`\nu F_{\nu}`. The default is ``False``.
        """
        params = dict(params) if params is not None else {}
        if params:
            if any(np.shape(v) != () for v in params.values()):
                raise ValueError('params must be scalars')
        egrid = dict(egrid) if egrid is not None else {}
        config = self.config
        colors = self.colors
        cl = config.cl
        comps = config.plot_comps
        step_kwargs = {'lw': 1.618, 'alpha': config.alpha}
        ribbon_kwargs = {'lw': 0.618, 'alpha': 0.2 * config.alpha}

        if mtype == 'ne':
            label_type = 'ne'
        elif mtype == 'ene':
            label_type = 'Fv' if label_Fv else 'ene'
        elif mtype == 'eene':
            label_type = 'vFv' if label_vFv else 'eene'
        else:
            raise ValueError("mtype must be 'ne', 'ene', or 'eene'")

        ax.set_ylabel(config._YLABLES[label_type])

        for name, data in self.data.items():
            color = colors[name]
            egrid_ = egrid.get(name, data.photon_egrid)
            ne, ci = data.unfolded_model(mtype, egrid_, params, False, cl)
            _plot_step(
                ax, egrid_[:-1], egrid_[1:], ne, color=color, **step_kwargs
            )
            if ci is not None:
                _plot_ribbon(
                    ax,
                    egrid_[:-1],
                    egrid_[1:],
                    ci,
                    color=color,
                    **ribbon_kwargs,
                )

            if comps:
                if not data.has_comps:
                    continue

                ne, ci = data.unfolded_model(mtype, egrid_, params, True)
                for ne_ in ne.values():
                    _plot_step(
                        ax,
                        egrid_[:-1],
                        egrid_[1:],
                        ne_,
                        color=color,
                        **(step_kwargs | {'ls': ':'}),
                    )
                if ci is not None:
                    for ci_ in ci.values():
                        _plot_ribbon(
                            ax,
                            egrid_[:-1],
                            egrid_[1:],
                            ci_,
                            color=color,
                            **ribbon_kwargs,
                        )

    def plot_folded(self, ax: Axes):
        """Plot folded model.

        Parameters
        ----------
        ax : Axes
            The Axes object to plot.
        """
        config = self.config
        colors = self.colors
        cl = config.cl
        step_kwargs = {'lw': 1.618, 'alpha': config.alpha}
        ribbon_kwargs = {'lw': 0.618, 'alpha': 0.2 * config.alpha}

        ax.set_ylabel(config._YLABLES['ce'])

        for name, data in self.data.items():
            color = colors[name]

            _plot_step(
                ax,
                data.channel_emin,
                data.channel_emax,
                data.ce_model,
                color=color,
                **step_kwargs,
            )

            quantiles = []
            for i_cl in cl:
                if (q := data.ce_model_ci(i_cl)) is not None:
                    quantiles.append(q)

            if quantiles:
                _plot_ribbon(
                    ax,
                    data.channel_emin,
                    data.channel_emax,
                    quantiles,
                    color=color,
                    **ribbon_kwargs,
                )

    def plot_ce(self, ax: Axes):
        """Plot data.

        Parameters
        ----------
        ax : Axes
            The Axes object to plot.
        """
        config = self.config
        colors = self.colors
        alpha = config.alpha
        xlog = config.xscale == 'log'

        ax.set_ylabel(config._YLABLES['ce'])

        for name, data in self.data.items():
            color = colors[name]
            marker = self._markers[name]
            ax.errorbar(
                x=data.channel_emean if xlog else data.channel_emid,
                xerr=data.channel_errors if xlog else 0.5 * data.channel_width,
                y=data.ce_data,
                yerr=data.ce_errors,
                alpha=alpha,
                color=color,
                fmt=f'{marker} ',
                label=name,
                lw=0.75,
                ms=2.4,
                mec=color,
                mfc='#FFFFFFCC',
            )

        if len(self.data) > 5:
            ncols = int(np.ceil(len(self.data) / 4))
        else:
            ncols = 1
        ax.legend(ncols=ncols)

    def plot_residuals(
        self,
        ax: Axes,
        rtype: Literal['rd', 'rp', 'rq'] | None = None,
    ):
        """Plot residuals.

        Parameters
        ----------
        ax : Axes
            The Axes object to plot.
        rtype : {'rd', 'rp', 'rq'}, optional
            The type of residuals, available options are:

                * ``'rd'``: deviance residuals
                * ``'rp'``: Pearson residuals
                * ``'rq'``: quantile residuals

        """
        if rtype not in {'rd', 'rp', 'rq', None}:
            raise ValueError(
                'residuals type must be "rd" (deviance), "rp" (pearson), '
                '"rq" (quantile), or None (use default residuals)'
            )

        config = self.config
        colors = self.colors
        cl = config.cl
        random_quantile = config.random_quantile
        with_sign = config.residuals_ci_with_sign
        mark_outlier = config.mark_outlier_residuals
        seed = config.seed
        ribbon_kwargs = {'lw': 0.618, 'alpha': 0.15 * config.alpha}

        if rtype is None:
            rtype = config.residuals

        alpha = config.alpha
        xlog = config.xscale == 'log'

        normal_q = stats.norm.isf(0.5 * (1.0 - cl))

        ax.set_ylabel(config._YLABLES[rtype])

        for name, data in self.data.items():
            color = colors[name]
            marker = self._markers[name]
            x = data.channel_emean if xlog else data.channel_emid
            xerr = data.channel_errors if xlog else 0.5 * data.channel_width

            quantiles = []
            for i_cl in cl:
                q = data.residuals_ci(
                    rtype, i_cl, seed, random_quantile, with_sign
                )
                if q is not None:
                    quantiles.append(q)

            if self.config.fill_residuals_ci:
                if quantiles:
                    _plot_ribbon(
                        ax,
                        data.channel_emin,
                        data.channel_emax,
                        quantiles,
                        color=color,
                        **ribbon_kwargs,
                    )
                else:
                    for q in normal_q:
                        ax.fill_between(
                            [data.channel_emin[0], data.channel_emax[-1]],
                            -q,
                            q,
                            color=color,
                            **ribbon_kwargs,
                        )

            use_mle = True if quantiles else False
            r = data.residuals(rtype, seed, config.random_quantile, use_mle)
            if rtype == 'rq':
                r, lower, upper = r
            else:
                lower = upper = False
            ax.errorbar(
                x=x,
                y=r,
                yerr=1.0,
                xerr=xerr,
                color=color,
                alpha=alpha,
                linewidth=0.75,
                linestyle='',
                marker=marker,
                markersize=2.4,
                markeredgecolor=color,
                markerfacecolor='#FFFFFFCC',
                lolims=lower,
                uplims=upper,
            )

            if mark_outlier:
                if quantiles:
                    q = quantiles[-1]
                else:
                    q = [-normal_q[-1], normal_q[-1]]
                mask = (r < q[0]) | (r > q[1])
                ax.scatter(x[mask], r[mask], marker='x', c='r')

        for q in normal_q:
            ax.axhline(q, ls=':', lw=1, c='gray', zorder=0)
            ax.axhline(-q, ls=':', lw=1, c='gray', zorder=0)

        ax.axhline(0, ls='--', lw=1, c='gray', zorder=0)
        yabs_max = abs(max(ax.get_ylim(), key=abs))
        ax.set_ylim(ymin=-yabs_max, ymax=yabs_max)

    def plot_qq(
        self,
        rtype: Literal['rd', 'rp', 'rq'] | None = None,
        seed: int | None = None,
        detrend: bool = True,
    ) -> Figure:
        """Quantile-Quantile plot.

        Parameters
        ----------
        rtype : {'rd', 'rp', 'rq'}, optional
            The type of residuals, available options are:

                * ``'rd'``: deviance residuals
                * ``'rp'``: Pearson residuals
                * ``'rq'``: quantile residuals
                * ``None``: use the default residuals type

            The default is ``None``.
        seed : int, optional
            Random seed used in calculation. The default is ``None``.
        detrend : bool, optional
            Whether to detrend the Q-Q plot. The default is ``True``.

        Returns
        -------
        Figure
            The Figure object containing Q-Q plot.
        """
        config = self.config
        random_quantile = config.random_quantile
        if rtype is None:
            rtype = config.residuals

        rsim = {
            name: data.residuals_sim(rtype, seed, random_quantile)
            for name, data in self.data.items()
        }
        if any(i is None for i in rsim.values()):
            rsim['total'] = None
        else:
            rsim['total'] = np.hstack(list(rsim.values()))

        use_mle = True if rsim['total'] is not None else False
        r = {
            name: data.residuals(rtype, seed, random_quantile, use_mle)
            for name, data in self.data.items()
        }
        if rtype == 'rq':
            r = {k: v[0] for k, v in r.items()}
        r['total'] = np.hstack(list(r.values()))

        n_subplots = len(self.data)
        if n_subplots == 1:
            ncols = 1
        else:
            ncols = n_subplots // 2
            if n_subplots % 2:
                ncols += 1

        fig = plt.figure(figsize=(4 + ncols * 2.25, 4), tight_layout=True)
        gs1 = fig.add_gridspec(1, 2, width_ratios=[4, ncols * 2.25])
        gs2 = gs1[0, 1].subgridspec(2, ncols, wspace=0.35)
        ax1 = fig.add_subplot(gs1[0, 0])
        ax2 = gs2.subplots(squeeze=False).ravel()
        if n_subplots % 2:
            ax2[-1].set_visible(False)
        ax2 = ax2[: len(self.data)]
        ax1.set_xlabel('Normal Theoretical Quantiles')
        ax1.set_ylabel('Residuals')

        alpha = config.alpha
        ha = 'center' if detrend else 'left'
        text_x = 0.5 if detrend else 0.03

        axs = {'total': ax1} | dict(zip(self.data.keys(), ax2, strict=True))
        colors = {'total': 'k'} | self.colors
        for name, ax in axs.items():
            color = colors[name]
            theor, q, line, lo, up = _get_qq(
                r[name], detrend, 0.95, rsim[name]
            )
            ax.scatter(theor, q, s=5, color=color, alpha=alpha)
            ax.plot(theor, line, ls='--', color=color, alpha=alpha)
            ax.plot(theor, lo, ls=':', color=color, alpha=alpha)
            ax.plot(theor, up, ls=':', color=color, alpha=alpha)
            ax.fill_between(
                theor, lo, up, alpha=0.2 * alpha, color=color, lw=0.0
            )
            ax.annotate(
                name,
                xy=(text_x, 0.97),
                xycoords='axes fraction',
                ha=ha,
                va='top',
                color=color,
            )

        return fig

    def plot_pit(self, detrend=True) -> Figure:
        """Probability integral transformation empirical CDF plot.

        Parameters
        ----------
        detrend : bool, optional
            Whether to detrend the PIT ECDF plot. The default is ``True``.

        Returns
        -------
        Figure
            The Figure object containing PIT ECDF plot.
        """
        config = self.config

        pit = {name: data.pit()[1] for name, data in self.data.items()}
        pit['total'] = np.hstack(list(pit.values()))

        n_subplots = len(self.data)
        if n_subplots == 1:
            ncols = 1
        else:
            ncols = n_subplots // 2
            if n_subplots % 2:
                ncols += 1

        fig = plt.figure(figsize=(4 + ncols * 2.25, 4), tight_layout=True)
        gs1 = fig.add_gridspec(1, 2, width_ratios=[4, ncols * 2.25])
        gs2 = gs1[0, 1].subgridspec(2, ncols, wspace=0.35)
        ax1 = fig.add_subplot(gs1[0, 0])
        ax2 = gs2.subplots(squeeze=False).ravel()
        if n_subplots % 2:
            ax2[-1].set_visible(False)
        ax2 = ax2[: len(self.data)]
        ax1.set_xlabel('PIT Value')
        ax1.set_ylabel('ECDF')

        alpha = config.alpha
        ha = 'right' if detrend else 'left'
        text_x = 0.97 if detrend else 0.03

        axs = {'total': ax1} | dict(zip(self.data.keys(), ax2, strict=True))
        colors = {'total': 'k'} | self.colors
        for name, ax in axs.items():
            color = colors[name]
            x, y, line, lower, upper = _get_pit_ecdf(pit[name], 0.95, detrend)
            ax.plot(x, line, ls='--', color=color, alpha=alpha)
            ax.fill_between(
                x, lower, upper, alpha=0.2 * alpha, color=color, step='mid'
            )
            ax.step(x, y, alpha=alpha, color=color, where='mid')
            ax.annotate(
                text=name,
                xy=(text_x, 0.97),
                xycoords='axes fraction',
                ha=ha,
                va='top',
                color=color,
            )

        return fig

    def plot_gof(self) -> Figure:
        """Plot distribution of GOF statistics and p-value.

        Returns
        -------
        Figure
            The Figure object containing GOF statistics plot.
        """
        if isinstance(self, MLEResultPlotter):
            if self._result._boot is None:
                raise RuntimeError(
                    'MLEResult.boot() must be called to assess gof'
                )
            n = int(self._result._boot.n_valid)
            dev_obs = self._result.deviance
            dev_sim = self._result._boot.deviance
            dev_sim = dev_sim['group'] | {'total': dev_sim['total']}
            p_value = self._result._boot.p_value
        elif isinstance(self, PosteriorResultPlotter):
            if self._result._ppc is None:
                raise RuntimeError(
                    'PosteriorResult.ppc() must be called to assess gof'
                )
            n = int(self._result._ppc.n_valid)
            dev_obs = self._result._mle['deviance']
            dev_sim = self._result._ppc.deviance
            dev_obs = dev_obs['group'] | {'total': dev_obs['total']}
            dev_sim = dev_sim['group'] | {'total': dev_sim['total']}
            p_value = self._result._ppc.p_value
        else:
            raise NotImplementedError
        p_value = p_value['group'] | {'total': p_value['total']}

        n_subplots = len(self.data)
        if n_subplots == 1:
            ncols = 1
        else:
            ncols = n_subplots // 2
            if n_subplots % 2:
                ncols += 1

        fig = plt.figure(figsize=(4 + ncols * 2.25, 4), tight_layout=True)
        gs1 = fig.add_gridspec(1, 2, width_ratios=[4, ncols * 2.25])
        gs2 = gs1[0, 1].subgridspec(2, ncols, wspace=0.35)
        ax1 = fig.add_subplot(gs1[0, 0])
        ax2 = gs2.subplots(squeeze=False).ravel()
        if n_subplots % 2:
            ax2[-1].set_visible(False)
        ax2 = ax2[: len(self.data)]
        ax1.set_xlabel('$D$')
        ax1.set_ylabel(r'$P(\mathcal{D} \geq D)$')

        axs = {'total': ax1} | dict(zip(self.data.keys(), ax2, strict=True))
        colors = {'total': 'k'} | self.colors
        for name, ax in axs.items():
            color = colors[name]
            d_obs = dev_obs[name]
            d_sim = np.sort(dev_sim[name])
            sf = 1.0 - np.arange(1.0, n + 1.0) / n
            ax.plot(d_sim, sf, color=color)
            ax.axvline(d_obs, color=color, ls=':')
            p = p_value[name]
            if p > 0.0:
                pstr = f'{name} $p = {p:.2g}$'
            else:
                pstr = f'{name} $p < {1}/{n}$'
            ax.annotate(
                text=pstr,
                xy=(0.97, 0.97),
                xycoords='axes fraction',
                ha='right',
                va='top',
                color=color,
            )
            ax.set_yscale('log')

        return fig


class MLEResultPlotter(Plotter):
    data: dict[str, MLEPlotData]
    _result: MLEResult
    _supported = (
        'data',
        'r',
        'rd',
        'rp',
        'rq',
        'ne',
        'ene',
        'eene',
        'Fv',
        'vFv',
        'corner',
        'gof',
        'qq',
        'pit',
    )

    def __call__(self, plots: str = 'data ne r') -> dict[str, Figure]:
        r"""Plot MLE fit results.

        Parameters
        ----------
        plots : str, optional
            Plots to show, available plots are:

                * ``'data'``: data and folded model plot
                * ``'ne'``: :math:`N(E)` model plot
                * ``'ene'``: :math:`E N(E)` model plot
                * ``'eene'``: :math:`E^2 N(E)` model plot
                * ``'Fv'``: :math:`F_{\nu}` model plot
                * ``'vFv'``: :math:`\nu F_{\nu}` model plot
                * ``'r'``: default residuals plot
                * ``'rd'``: deviance residuals plot
                * ``'rp'``: Pearson residuals plot
                * ``'rq'``: quantile residuals plot
                * ``'corner'``: corner plot
                * ``'gof'``: goodness-of-fit statistics plot
                * ``'qq'``: quantiles-quantiles plot of residuals
                * ``'pit'``: probability integral transform plot of spectral
                  data

            Multiple plots can be combined by separating them with whitespace.
            THe default is ``'data ne r'``.

        Returns
        -------
        dict
            Dictionary containing Figure object for each plot.
        """
        plots = re.split(r'\s+', str(plots))
        if any(p not in self._supported for p in plots):
            supported = ', '.join(self._supported)
            err = ', '.join(p for p in plots if p not in self._supported)
            raise ValueError(f'supported plots are: {supported}; got {err}')

        plots_set = set(plots)
        dic = {}

        spec = {
            'data',
            'ne',
            'ene',
            'eene',
            'Fv',
            'vFv',
            'r',
            'rd',
            'rp',
            'rq',
        }
        if spec & plots_set:
            residuals = [i for i in plots if i in ('r', 'rd', 'rp', 'rq')]
            if residuals:
                residuals = residuals[-1]
                if residuals == 'r':
                    residuals = True
            else:
                residuals = False
            dic['spec'] = self.plot_spec(
                data='data' in plots,
                ne='ne' in plots,
                ene=bool({'ene', 'Fv'} & plots_set),
                eene=bool({'eene', 'vFv'} & plots_set),
                residuals=residuals,
                label_Fv='Fv' in plots,
                label_vFv='vFv' in plots,
            )

        if 'corner' in plots_set:
            dic['corner'] = self.plot_corner()

        if 'gof' in plots_set:
            dic['gof'] = self.plot_gof()

        if 'qq' in plots_set:
            dic['qq'] = self.plot_qq()

        if 'pit' in plots_set:
            dic['pit'] = self.plot_pit()

        return dic

    @staticmethod
    def get_plot_data(result: MLEResult) -> dict[str, MLEPlotData]:
        helper = result._helper
        keys = jax.random.split(
            jax.random.PRNGKey(helper.seed['resd']), len(helper.data_names)
        )
        data = {
            name: MLEPlotData(name, result, int(key[0]))
            for name, key in zip(helper.data_names, keys, strict=True)
        }
        return data

    def plot_corner(
        self,
        params: str | Sequence[str] | None = None,
        color: str | None = None,
        bins: int | Sequence[int] = 40,
        hist_bin_factor: float | Sequence[float] = 1.5,
        fig_path: str | None = None,
    ) -> Figure:
        """Corner plot of bootstrap parameters.

        Parameters
        ----------
        params : str or sequence of str, optional
            Parameters to plot. The default is all spectral parameters.
        color : str, optional
            Color of the plot. The default is ``None``.
        bins : int or list of int, optional
            The number of bins to use in histograms, either as a fixed value
            for all dimensions or as a list of integers for each dimension.
            The default is 40.
        hist_bin_factor : float or list of float, optional
            This is a factor (or list of factors, one for each dimension)
            that will multiply the bin specifications when making the 1-D
            histograms. This is generally used to increase the number of
            bins in the 1-D plots to provide more resolution.
            The default is 1.5.
        fig_path : str, optional
            Path to save the figure. The default is ``None``.

        Returns
        -------
        Figure
            The Figure object containing corner plot.
        """
        if self._result._boot is None:
            raise ValueError('MLEResult.boot() must be called first')

        helper = self._result._helper
        params = check_params(params, helper)
        axes_scale = [
            'log' if helper.params_log[p] else 'linear' for p in params
        ]
        params_titles = self.params_titles
        params_labels = self.params_labels
        fig = plot_corner(
            idata=az.from_dict(self._result._boot.params),
            bins=bins,
            hist_bin_factor=hist_bin_factor,
            params=params,
            axes_scale=axes_scale,
            levels=self.config.cl,
            titles=[params_titles[p] for p in params],
            labels=[params_labels[p] for p in params],
            color=color,
        )
        if fig_path:
            fig.savefig(fig_path, bbox_inches='tight')
        return fig


class PosteriorResultPlotter(Plotter):
    data: dict[str, PosteriorPlotData]
    _result: PosteriorResult
    _supported = (
        'data',
        'r',
        'rd',
        'rp',
        'rq',
        'ne',
        'ene',
        'eene',
        'Fv',
        'vFv',
        'corner',
        'gof',
        'qq',
        'pit',
        'trace',
        'khat',
    )

    def __call__(self, plots: str = 'data ne r') -> dict[str, Figure]:
        r"""Plot Bayesian fit results.

        Parameters
        ----------
        plots : str, optional
            Plots to show, available plots are:

                * ``'data'``: data and folded model plot
                * ``'ne'``: :math:`N(E)` model plot
                * ``'ene'``: :math:`E N(E)` model plot
                * ``'eene'``: :math:`E^2 N(E)` model plot
                * ``'Fv'``: :math:`F_{\nu}` model plot
                * ``'vFv'``: :math:`\nu F_{\nu}` model plot
                * ``'r'``: default residuals plot
                * ``'rd'``: deviance residuals plot
                * ``'rp'``: Pearson residuals plot
                * ``'rq'``: PSIS-LOO quantile residuals plot
                * ``'corner'``: corner plot
                * ``'gof'``: goodness-of-fit statistics plot
                * ``'qq'``: quantiles-quantiles plot of residuals
                * ``'pit'``: PSIS-LOO probability integral transform plot of
                  spectral data
                * ``'trace'``: trace plot of posterior samples
                * ``'khat'``: k-hat plot for Bayesian PSIS-LOO diagnostics

            Multiple plots can be combined by separating them with whitespace.
            THe default is ``'data ne r'``.

        Returns
        -------
        dict
            Dictionary containing Figure object for each plot.
        """
        plots = re.split(r'\s+', str(plots))
        if any(p not in self._supported for p in plots):
            supported = ', '.join(self._supported)
            err = ', '.join(p for p in plots if p not in self._supported)
            raise ValueError(f'supported plots are: {supported}; got {err}')

        plots_set = set(plots)
        dic = {}

        spec = {
            'data',
            'ne',
            'ene',
            'eene',
            'Fv',
            'vFv',
            'r',
            'rd',
            'rp',
            'rq',
        }
        if spec & plots_set:
            residuals = [i for i in plots if i in ('r', 'rd', 'rp', 'rq')]
            if residuals:
                residuals = residuals[-1]
                if residuals == 'r':
                    residuals = True
            else:
                residuals = False
            dic['spec'] = self.plot_spec(
                data='data' in plots,
                ne='ne' in plots,
                ene=bool({'ene', 'Fv'} & plots_set),
                eene=bool({'eene', 'vFv'} & plots_set),
                residuals=residuals,
                label_Fv='Fv' in plots,
                label_vFv='vFv' in plots,
            )

        if 'corner' in plots_set:
            dic['corner'] = self.plot_corner()

        if 'gof' in plots_set:
            dic['gof'] = self.plot_gof()

        if 'qq' in plots_set:
            dic['qq'] = self.plot_qq()

        if 'pit' in plots_set:
            dic['pit'] = self.plot_pit()

        if 'trace' in plots_set:
            dic['trace'] = self.plot_trace()

        if 'khat' in plots_set:
            dic['khat'] = self.plot_khat()

        return dic

    @staticmethod
    def get_plot_data(result: PosteriorResult) -> dict[str, PosteriorPlotData]:
        helper = result._helper
        keys = jax.random.split(
            jax.random.PRNGKey(helper.seed['resd']), len(helper.data_names)
        )
        data = {
            name: PosteriorPlotData(name, result, int(key[0]))
            for name, key in zip(helper.data_names, keys, strict=True)
        }
        return data

    def plot_trace(
        self,
        params: str | Sequence[str] | None = None,
        fig_path: str | None = None,
    ) -> Figure:
        """Plot trace plot of posterior samples.

        Parameters
        ----------
        params : str or sequence of str, optional
            Parameters to plot. The default is all spectral parameters.
        fig_path : str, optional
            Path to save the figure. The default is ``None``.
        """
        helper = self._result._helper
        params = check_params(params, helper)
        axes_scale = [
            'log' if helper.params_log[p] else 'linear' for p in params
        ]
        params_labels = self.params_labels
        labels = [params_labels[p] for p in params]
        fig = plot_trace(self._result._idata, params, axes_scale, labels)
        if fig_path:
            fig.savefig(fig_path, bbox_inches='tight')
        return fig

    def plot_corner(
        self,
        params: str | Sequence[str] | None = None,
        color: str | None = None,
        divergences: bool = True,
        bins: int | Sequence[int] = 40,
        hist_bin_factor: float | Sequence[float] = 1.5,
        fig_path: str | None = None,
    ) -> Figure:
        """Corner plot of posterior parameters.

        Parameters
        ----------
        params : str or sequence of str, optional
            Parameters to plot. The default is all spectral parameters.
        color : str, optional
            Color of the plot. The default is ``None``.
        divergences : bool, optional
            Whether to show divergent samples. The default is ``True``.
        bins : int or list of int, optional
            The number of bins to use in histograms, either as a fixed value
            for all dimensions, or as a list of integers for each dimension.
            The default is 40.
        hist_bin_factor : float or list of float, optional
            This is a factor (or list of factors, one for each dimension)
            that will multiply the bin specifications when making the 1-D
            histograms. This is generally used to increase the number of
            bins in the 1-D plots to provide more resolution.
            The default is 1.5.
        fig_path : str, optional
            Path to save the figure. The default is ``None``.

        Returns
        -------
        Figure
            The Figure object containing corner plot.
        """
        helper = self._result._helper
        params = check_params(params, helper)
        axes_scale = [
            'log' if helper.params_log[p] else 'linear' for p in params
        ]
        params_titles = self.params_titles
        params_labels = self.params_labels
        fig = plot_corner(
            idata=self._result._idata,
            bins=bins,
            hist_bin_factor=hist_bin_factor,
            params=params,
            axes_scale=axes_scale,
            levels=self.config.cl,
            titles=[params_titles[p] for p in params],
            labels=[params_labels[p] for p in params],
            color=color,
            divergences=divergences,
        )
        if fig_path:
            fig.savefig(fig_path, bbox_inches='tight')
        return fig

    def plot_khat(self) -> Figure:
        """Plot k-hat diagnostic of PSIS-LOO."""
        config = self.config
        colors = self.colors
        alpha = config.alpha
        xlog = config.xscale == 'log'

        fig, ax = plt.subplots(1, 1, squeeze=True, tight_layout=True)

        khat = self._result.loo.pareto_k
        if np.any(khat.values > 0.7):
            ax.axhline(0.7, color='r', lw=0.5, ls=':')

        for name, data in self.data.items():
            color = colors[name]
            marker = self._markers[name]
            khat_data = khat.sel(channel=data.channel).values
            x = data.channel_emean if xlog else data.channel_emid
            ax.errorbar(
                x=x,
                xerr=data.channel_errors if xlog else 0.5 * data.channel_width,
                y=khat_data,
                alpha=alpha,
                color=color,
                fmt=f'{marker} ',
                label=name,
                lw=0.75,
                ms=2.4,
                mec=color,
                mfc='#FFFFFFCC',
            )

            mask = khat_data > 0.7
            if np.any(mask):
                ax.scatter(x=x[mask], y=khat_data[mask], marker='x', c='r')

        ax.set_xscale(config.xscale)
        ax.set_xlabel('Energy [keV]')
        ax.set_ylabel(r'Shape Parameter $\hat{k}$')

        return fig



"""Helper for fitting and analysis."""

from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, NamedTuple

import jax
import jax.numpy as jnp
import numpy as np
import numpyro
import optimistix as optx
from jax import lax
from numpyro import handlers
from numpyro.infer.util import constrain_fn, unconstrain_fn

from elisa.infer.likelihood import (
    _STATISTIC_BACK_NORMAL,
    _STATISTIC_SPEC_NORMAL,
    _STATISTIC_WITH_BACK,
    chi2,
    cstat,
    pgstat,
    pstat,
    wstat,
)
from elisa.util.config import get_parallel_number
from elisa.util.misc import (
    get_unit_latex,
    progress_bar_factory,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Literal

    from numpyro.distributions import Distribution

    from elisa.data.base import FixedData
    from elisa.infer.fit import Fit
    from elisa.infer.likelihood import Statistic
    from elisa.models.model import CompiledModel, ModelInfo, ParamSetup
    from elisa.util.typing import (
        JAXArray,
        JAXFloat,
        ParamID,
        ParamName,
        ParamNameValMapping,
    )


def check_params(
    params: str | Sequence[str] | None, helper: Helper
) -> list[str]:
    params_names = helper.params_names

    all_params = set(params_names['all']) | set(helper.params_setup)
    forwarded = {
        k: v[0]
        for k, v in helper.params_setup.items()
        if v[1].name == 'Forwarded'
    }
    fixed = [k for k, v in helper.params_setup.items() if v[1].name == 'Fixed']
    integrated = [
        k for k, v in helper.params_setup.items() if v[1].name == 'Integrated'
    ]

    if params is None:
        params = set(params_names['interest'])

    elif isinstance(params, str):
        # check if params exist
        if params not in all_params:
            raise ValueError(f'parameter {params} is not exist')

        params = {params}

    elif isinstance(params, Iterable):
        # check if params exist
        params = {str(i) for i in params}
        if not params.issubset(all_params):
            params_err = params - set(params_names['all'])
            raise ValueError(f'parameters: {params_err} are not exist')

    else:
        raise ValueError('params must be str, or sequence of str')

    if params_err := params.intersection(forwarded):
        forwarded = {i: forwarded[i] for i in params_err}
        info = ', '.join(f'{k} to {v}' for k, v in forwarded.items())
        raise RuntimeError(
            f"parameters are linked: {info}; corresponding parameters' "
            'name should be used'
        )

    if params_err := params.intersection(fixed):
        info = ', '.join(params_err)
        raise RuntimeError(f'parameters are fixed: {info}')

    if params_err := params.intersection(integrated):
        info = ', '.join(params_err)
        raise RuntimeError(f'parameters are integrated-out: {info}')

    return sorted(params, key=params_names['all'].index)


# def get_reparam(dist: Distribution) -> tuple[Reparam, Callable] | None:
#     """Get reparam for a distribution."""
#     # TODO: support more reparameterizers
#     if isinstance(dist, (Normal, StudentT, Cauchy)):
#         param = '_centered'
#         suffix = '_decentered'
#         reparam = LocScaleReparam()
#     elif isinstance(dist, TransformedDistribution):
#         suffix = '_base'
#         reparam = TransformReparam()
#         transforms = dist.transforms
#         def inv(x):
#             """Inverse transformation."""
#             for t in reversed(transforms):
#                 x = t.inv(x)
#             return x
#     elif isinstance(dist, ProjectedNormal):
#         suffix = '_normal'
#         reparam = ProjectedNormalReparam()
#     elif isinstance(dist, VonMises):
#         suffix = '_unwrapped'
#         reparam = CircularReparam()
#     else:
#         return None
#
#     return reparam, inv


def get_helper(fit: Fit) -> Helper:
    """Get helper functions for fitting."""
    model_info: ModelInfo = fit._model_info
    data: dict[str, FixedData] = fit._data
    model: dict[str, CompiledModel] = fit._model
    stat: dict[str, Statistic] = fit._stat
    seed0 = fit._seed
    rng_seed: dict[str, int] = {
        'mcmc': seed0,  # for MCMC
        'pred': seed0 + 1,  # for data simulation
        'resd': seed0 + 2,  # for random quantile residuals
    }

    # channel number of data
    ndata = {k: v.channel.size for k, v in data.items()}
    ndata['total'] = sum(ndata.values())

    # channel information
    channels = {f'{k}_channel': v.channel for k, v in data.items()}
    channels['channel'] = np.hstack(list(channels.values()))

    # number of free parameters
    nparam = len(model_info.sample)

    # degree of freedom
    dof = ndata['total'] - nparam

    # ======================== count data calculator ==========================
    on_names = [f'{i}_Non' for i in data]
    off_names = [f'{i}_Noff' for i in data if stat[i] in _STATISTIC_WITH_BACK]
    back_ratio = {
        k: v.back_ratio
        for k, v in data.items()
        if stat[k] in _STATISTIC_WITH_BACK
    }
    spec_unit = {
        k: 1.0 / (v.spec_exposure * v.channel_width) for k, v in data.items()
    }

    @jax.jit
    def get_counts_data(counts: dict[str, JAXArray]) -> dict[str, JAXArray]:
        """Get count data, including "on", "off" and net counts."""
        counts = {k: jnp.asarray(v, float) for k, v in counts.items()}
        assert set(counts) == set(on_names + off_names)

        # counts in the "on" measurement of each dataset
        counts_data = {i: counts[i] for i in on_names}

        # counts in the "off" measurement of each dataset
        counts_data |= {i: counts[i] for i in off_names}

        net_counts = {
            i: (
                counts[f'{i}_Non']
                - back_ratio.get(i, 0.0) * counts.get(f'{i}_Noff', 0.0)
            )
            for i in data
        }

        # net spectrum [counts s^-1 keV^-1]
        counts_data |= {i: net_counts[i] * spec_unit[i] for i in data.keys()}

        # stack net spectrum of all channels of all datasets
        counts_data['channels'] = jnp.concatenate(
            [counts_data[i] for i in data.keys()],
            axis=-1,
        )

        # total net counts of datasets
        counts_data['total'] = jnp.sum(
            jnp.asarray([i.sum(axis=-1) for i in net_counts.values()]), axis=0
        )

        return counts_data

    # ======================== count data calculator ==========================

    obs_counts = {
        f'{k}_Non': (
            v.net_counts
            if stat[k] in _STATISTIC_SPEC_NORMAL
            else v.spec_counts
        )
        for k, v in data.items()
    }
    obs_counts |= {
        f'{k}_Noff': v.back_counts
        for k, v in data.items()
        if stat[k] in _STATISTIC_WITH_BACK
    }
    obs_data = get_counts_data(obs_counts)

    # ======================== count data simulator ===========================
    def simulator_factory(data_dist: Literal['norm', 'poisson'], *dist_args):
        """Factory to create data simulator."""

        def simulator(
            rng: np.random.Generator,
            model_values: np.ndarray,
            n: int,
        ):
            """Simulate data given random number generator and model values."""
            if n != 1:
                shape = (n,) + model_values.shape
            else:
                shape = model_values.shape

            if data_dist == 'norm':
                # TODO: fix the negative counts by setting them to zeros
                return rng.normal(model_values, *dist_args, shape)
            elif data_dist == 'poisson':
                return rng.poisson(model_values, shape)
            else:
                raise NotImplementedError(f'{data_dist = }')

        return simulator

    simulators = {}
    sampling_dist: dict[str, tuple[Literal['norm', 'poisson'], tuple]] = {}
    for k, s in stat.items():
        d = data[k]

        name = f'{k}_Non'
        if s in _STATISTIC_SPEC_NORMAL:
            simulators[name] = simulator_factory('norm', d.spec_errors)
            sampling_dist[name] = ('norm', (d.spec_errors,))
        else:
            simulators[name] = simulator_factory('poisson')
            sampling_dist[name] = ('poisson', ())

        if s in _STATISTIC_WITH_BACK:
            name = f'{k}_Noff'
            if s in _STATISTIC_BACK_NORMAL:
                simulators[name] = simulator_factory('norm', d.back_errors)
                sampling_dist[name] = ('norm', (d.spec_errors,))
            else:
                simulators[name] = simulator_factory('poisson')
                sampling_dist[name] = ('poisson', ())

    def simulate(
        rng_seed: int,
        model_values: dict[str, JAXArray],
        n: int = 1,
    ) -> dict[str, JAXArray]:
        """Simulate data given model values.
        Use numpy.random instead of numpyro.infer.Predictive for performance.
        """
        models = {i: model_values[f'{i}_model'] for i in simulators.keys()}
        rng = np.random.default_rng(int(rng_seed))
        sim = {k: v(rng, models[k], n) for k, v in simulators.items()}
        return get_counts_data(sim)

    # ======================== count data simulator ===========================

    # ======================== create numpyro model ===========================
    pname_to_latex: dict[ParamName, str] = {
        pname: model_info.latex[pid] for pid, pname in model_info.name.items()
    }
    pname_to_log: dict[ParamName, bool] = {
        pname: model_info.log[pid] for pid, pname in model_info.name.items()
    }

    pname_to_unit: dict[ParamName, str] = {
        pname: get_unit_latex(model_info.unit[pid], throw=False)
        for pid, pname in model_info.name.items()
    }
    pname_to_comp_latex: dict[ParamName, str] = {
        pname: model_info.pid_to_comp_latex[pid]
        for pid, pname in model_info.name.items()
    }

    # get model parameters priors
    pid_to_pname: dict[ParamID, ParamName] = model_info.name
    pname_to_pid: dict[ParamName, ParamID] = {
        v: k for k, v in pid_to_pname.items()
    }
    pid_to_prior: dict[ParamID, Distribution] = model_info.sample
    params_prior: dict[ParamName, Distribution] = {
        pid_to_pname[pid]: pid_to_prior[pid] for pid in pid_to_prior
    }

    # get deterministic value getter function
    deterministic: dict[ParamID, Callable] = model_info.deterministic

    # get the likelihood function for each dataset
    likelihood_wrapper = {
        'chi2': chi2,
        'cstat': cstat,
        'pstat': pstat,
        'wstat': wstat,
        'pgstat': pgstat,
    }
    likelihood: dict[str, Callable[[JAXArray], None]] = {
        k: likelihood_wrapper[stat[k]](v, model[k].eval)
        for k, v in data.items()
    }

    # get default re-parameterization of each parameter
    # reparams: dict[ParamName, tuple[Reparam, Callable]] = {
    #     name: reparam_and_inv
    #     for name, prior in params_prior.items  ()
    #     if (reparam_and_inv := get_reparam(prior)) is not None
    # }

    def numpyro_model(predictive: bool = False) -> None:
        """The numpyro model for spectral fitting."""
        # TODO:
        #  figure out how to handle reparam transformation, so we can
        #       * give initial parameter value in the original space and
        #         transformed to reparameterized space,
        #       * find the classic confidence interval in the reparameterized
        #         space and then transform back to the original space.
        #  This is not trivial because transformation is not always bijective!
        #  The trick to handle confidence interval of composite parameter may
        #  also be used here to solve the second problem.
        # with numpyro.handlers.reparam(config=reparams):
        #     params_name_values = {
        #         name: numpyro.sample(name, dist)
        #         for name, dist in params_prior.items()
        #     }

        # get parameter value from prior
        params_name_values = {
            name: numpyro.sample(name, dist)
            for name, dist in params_prior.items()
        }
        params_id_values = {
            pname_to_pid[name]: value
            for name, value in params_name_values.items()
        }

        # store composite parameters into chains
        for pid, fn in deterministic.items():
            numpyro.deterministic(pid_to_pname[pid], fn(params_id_values))

        # the likelihood between observation and model for each dataset
        jax.tree.map(
            lambda f: f(params_name_values, predictive=predictive),
            likelihood,
        )

    # ======================== create numpyro model ===========================

    # =================== functions used in optimization ======================
    params_names = [
        f'{i[1]}.{i[2]}' if i[1] else i[2] for i in model_info.info if i[0]
    ]
    interest_names = [f'{i[1]}.{i[2]}' for i in model_info.info if all(i[:2])]
    free_names = sorted(params_prior.keys(), key=params_names.index)
    deterministic_names = [pid_to_pname[i] for i in deterministic]
    deterministic_names = sorted(deterministic_names, key=params_names.index)

    # ensure if names are consistent
    if set(free_names + deterministic_names) != set(params_names):
        raise RuntimeError(
            f'{params_names = }, {free_names = }, {deterministic_names = }'
        )

    data_group = {
        k: (f'{k}_Non', f'{k}_Noff')
        if v in _STATISTIC_WITH_BACK
        else (f'{k}_Non',)
        for k, v in stat.items()
    }

    @jax.jit
    def arr_to_dic(arr: JAXArray) -> ParamNameValMapping:
        """Covert free parameters from an array to a dict."""
        assert len(arr) == len(free_names)
        return dict(zip(free_names, arr, strict=True))

    @jax.jit
    def dic_to_arr(dic: ParamNameValMapping) -> JAXArray:
        """Covert free parameters from a dict to an array."""
        return jnp.array([dic[i] for i in free_names], float)

    @jax.jit
    def unconstr_arr_to_constr_dic(arr: JAXArray) -> ParamNameValMapping:
        """Covert free parameters array from unconstrained space to dict in
        constrained space.
        """
        return constrain_fn(
            model=numpyro_model,
            model_args=(),
            model_kwargs={},
            params=arr_to_dic(arr),
        )

    @jax.jit
    def constr_arr_to_unconstr_dic(arr: JAXArray) -> ParamNameValMapping:
        """Covert free parameters array from constrained space to dict in
        unconstrained space.
        """
        return unconstrain_fn(
            model=numpyro_model,
            model_args=(),
            model_kwargs={},
            params=arr_to_dic(arr),
        )

    @jax.jit
    def constr_arr_to_unconstr_arr(arr: JAXArray) -> JAXArray:
        """Covert free parameters array from constrained space into
        unconstrained space.
        """
        unconstr_dic = constr_arr_to_unconstr_dic(arr)
        return jnp.asarray([unconstr_dic[i] for i in free_names])

    @jax.jit
    def constr_dic_to_unconstr_arr(dic: ParamNameValMapping) -> JAXArray:
        """Covert free parameters dict from constrained space to array in
        unconstrained space.
        """
        constr_arr = dic_to_arr(dic)
        return constr_arr_to_unconstr_arr(constr_arr)

    # get default value of each parameter
    default_constr_dic = {
        pid_to_pname[k]: v for k, v in model_info.default.items()
    }
    default_constr_dic = {k: default_constr_dic[k] for k in free_names}
    default_constr_arr = jnp.array([default_constr_dic[i] for i in free_names])
    default_unconstr_arr = constr_arr_to_unconstr_arr(default_constr_arr)
    free_default: dict[str, dict[ParamName, JAXFloat] | JAXArray] = {
        'constr_dic': default_constr_dic,
        'constr_arr': default_constr_arr,
        'unconstr_arr': default_unconstr_arr,
    }

    def get_sites(
        unconstr_arr: JAXArray,
    ) -> dict[
        Literal['params', 'models', 'loglike'], dict[str, JAXFloat | JAXArray]
    ]:
        """Get parameters in constrained space, models values and log
        likelihood, given free parameters array in unconstrained space.
        """
        sites = constrain_fn(
            model=numpyro_model,
            model_args=(),
            model_kwargs={},
            params=arr_to_dic(unconstr_arr),
            return_deterministic=True,
        )
        params = get_params(sites)
        models = get_models(sites)
        loglike = get_loglike(sites)
        return {'params': params, 'models': models, 'loglike': loglike}

    def get_params(sites: Mapping) -> dict:
        """Get parameters' values in constrained space given numpyro model
        sites.
        """
        params = {i: sites[i] for i in params_names}
        return params

    def get_models(sites: Mapping) -> dict:
        """Get model values given numpyro model sites."""
        models = {k: sites[k] for k in data_group.keys()}
        models |= {
            f'{i}_model': sites[f'{i}_model']
            for v in data_group.values()
            for i in v
        }
        return models

    def get_loglike(sites: Mapping) -> dict:
        """Get log likelihood given numpyro model sites."""
        loglike_data = {
            i: sites[f'{i}_loglike'] for v in data_group.values() for i in v
        }
        loglike_point = {i: sites[f'{i}_loglike'] for i in data_group.keys()}
        loglike_group = {k: v.sum(axis=-1) for k, v in loglike_point.items()}
        loglike_channels = jnp.concatenate(
            [loglike_point[i] for i in data_group.keys()], axis=-1
        )
        loglike_total = loglike_channels.sum(axis=-1)
        loglike = {
            'data': loglike_data,
            'point': loglike_point,
            'group': loglike_group,
            'channels': loglike_channels,
            'total': loglike_total,
        }
        return loglike

    @jax.jit
    def unconstr_dic_to_params_dic(
        dic: ParamNameValMapping,
    ) -> ParamNameValMapping:
        """Get parameters dict in constrained space,
        given a free parameters dict in unconstrained space.
        """
        return jax.jit(get_sites)(dic_to_arr(dic))['params']

    @jax.jit
    def unconstr_arr_to_params_array(arr: JAXArray) -> JAXArray:
        """Get parameters dict in constrained space,
        given a free parameters array in unconstrained space.
        """
        unconstr_dic = arr_to_dic(arr)
        params_dic = unconstr_dic_to_params_dic(unconstr_dic)
        return jnp.array([params_dic[i] for i in params_names])

    @jax.jit
    def unconstr_covar(unconstr_arr: JAXArray) -> JAXArray:
        """Calculate covariance matrix of free parameters in unconstrained
        space, given a free parameters array in unconstrained space.
        """
        hess = jax.jit(jax.hessian(deviance_total))(unconstr_arr)
        covar = jnp.linalg.inv(hess)
        return 2.0 * covar

    @jax.jit
    def params_covar(
        unconstr_arr: JAXArray,
        unconstr_cov: JAXArray,
    ) -> jnp.ndarray:
        """Calculate covariance matrix of all parameters in constrained space,
        given values and covariance matrix of free parameters in unconstrained
        space.
        """
        jac = jax.jit(jax.jacobian(unconstr_arr_to_params_array))(unconstr_arr)
        return jac @ unconstr_cov @ jac.T

    @jax.jit
    def get_mle(unconstr_arr: JAXArray) -> tuple[JAXArray, JAXArray]:
        """Get the value and covariance matrix of all parameters in constrained
        space, given MLE of free parameters in unconstrained space.
        """
        params_arr = unconstr_arr_to_params_array(unconstr_arr)
        params_cov = params_covar(unconstr_arr, unconstr_covar(unconstr_arr))
        return params_arr, params_cov

    # NOTE:
    #   the following functions will be used in simulation procedure,
    #   so we do not jit it here, or data substitution will fail
    def loglike(unconstr_arr: JAXArray) -> dict[str, JAXArray]:
        """Calculate log-likelihood given free parameters dict in constrained
        space.
        """
        return get_sites(unconstr_arr)['loglike']

    def deviance(unconstr_arr: JAXArray) -> dict:
        """Calculate total/group/point deviance given free parameters array in
        unconstrained space.
        """
        loglike_dic = loglike(unconstr_arr)
        neg_double = jax.jit(lambda x: -2.0 * x)
        point = jax.tree.map(neg_double, loglike_dic['point'])
        group = jax.tree.map(neg_double, loglike_dic['group'])
        total = jax.tree.map(neg_double, loglike_dic['total'])
        return {'total': total, 'group': group, 'point': point}

    def deviance_total(unconstr_arr: JAXArray) -> JAXFloat:
        """Calculate total deviance given free parameters array in
        unconstrained space.
        """
        return deviance(unconstr_arr)['total']

    def residual(unconstr_arr: JAXArray) -> JAXArray:
        """Calculate deviance residual (i.e. sqrt deviance) given free
        parameters array in unconstrained space.
        """
        loglike_dic = loglike(unconstr_arr)
        loglike_arr = jnp.hstack(list(loglike_dic['point'].values()))
        return jnp.sqrt(-2.0 * loglike_arr)

    # =================== functions used in optimization ======================

    # =============== functions used in simulation procedure ==================
    lm_solver = optx.LevenbergMarquardt(rtol=0.0, atol=1e-6)

    @jax.jit
    def fit_once(i: int, args: tuple) -> tuple:
        """Loop core, fit simulation data once."""
        result, init = args
        sim_data = result['data']

        # substitute observation data with simulation data
        new_data = {
            f'{j}_data': sim_data[j][i] for v in data_group.values() for j in v
        }
        new_residual = jax.jit(handlers.substitute(fn=residual, data=new_data))
        new_deviance = jax.jit(handlers.substitute(fn=deviance, data=new_data))
        new_sites = jax.jit(handlers.substitute(fn=get_sites, data=new_data))

        # fit simulation data
        res = optx.least_squares(
            fn=lambda p, _: new_residual(p),
            solver=lm_solver,
            y0=init[i],
            max_steps=1024,
            throw=False,
        )
        fitted_params = res.value
        grad_norm = jnp.linalg.norm(res.state.f_info.compute_grad())

        sites = new_sites(fitted_params)

        # update best fit params to result
        params = sites['params']
        result['params'] = jax.tree.map(
            lambda x, y: x.at[i].set(y),
            result['params'],
            params,
        )

        # update the best fit model to result
        models = sites['models']
        result['models'] = jax.tree.map(
            lambda x, y: x.at[i].set(y),
            result['models'],
            {k: models[k] for k in result['models']},
        )

        # update the deviance information to result
        dev = new_deviance(fitted_params)
        res_dev = result['deviance']
        res_dev['group'] = jax.tree.map(
            lambda x, y: x.at[i].set(y),
            res_dev['group'],
            dev['group'],
        )
        res_dev['point'] = jax.tree.map(
            lambda x, y: x.at[i].set(y),
            res_dev['point'],
            dev['point'],
        )
        res_dev['total'] = res_dev['total'].at[i].set(dev['total'])

        valid = jnp.bitwise_not(
            jnp.isnan(dev['total'])
            | jnp.isnan(grad_norm)
            | jnp.greater(grad_norm, 1e-3)
        )
        result['valid'] = result['valid'].at[i].set(valid)

        return result, init

    def fit_in_sequence(
        result: dict,
        init: JAXArray,
        run_str: str,
        progress: bool,
        update_rate: int,
    ):
        """Fit simulation data in sequence."""
        n = len(result['valid'])

        if progress:
            pbar_factory = progress_bar_factory(
                n, 1, run_str=run_str, update_rate=update_rate
            )
            fn = pbar_factory(fit_once)
        else:
            fn = fit_once

        fit_jit = jax.jit(lambda *args: lax.fori_loop(0, n, fn, args)[0])
        result = fit_jit(result, init)
        return result

    def fit_in_parallel(
        result: dict,
        init: JAXArray,
        run_str: str,
        progress: bool,
        update_rate: int,
        n_parallel: int,
    ) -> dict:
        """Fit simulation data in parallel."""
        n = len(result['valid'])
        n_parallel = int(n_parallel)
        batch = n // n_parallel

        if progress:
            pbar_factory = progress_bar_factory(
                n, n_parallel, run_str=run_str, update_rate=update_rate
            )
            fn = pbar_factory(fit_once)
        else:
            fn = fit_once

        fit_pmap = jax.pmap(lambda *args: lax.fori_loop(0, batch, fn, args)[0])
        reshape = lambda x: x.reshape((n_parallel, -1) + x.shape[1:])
        result = fit_pmap(
            jax.tree.map(reshape, result),
            jax.tree.map(reshape, init),
        )

        return jax.tree.map(jnp.concatenate, result)

    def batch_fit(
        init_params: dict[str, JAXArray],
        data: dict[str, JAXArray],
        parallel: bool = True,
        n_parallel: int | None = None,
        progress: bool = True,
        update_rate: int = 50,
        run_str: str = 'Fitting',
    ) -> dict:
        """Simulate data and then fit the simulation data.

        Parameters
        ----------
        init_params : dict
            The initial parameters values in constrained space.
        data : dict
            The simulation data corresponding to `free_params`.
        parallel : bool, optional
            Whether to fit in parallel, by default True.
        n_parallel : int, optional
            The number of parallel processes when `parallel` is ``True``.
            Defaults to ``jax.local_device_count()``.
        progress : bool, optional
            Whether to show progress bar, by default True.
        update_rate : int, optional
            The update rate of the progress bar, by default 50.
        run_str : str, optional
            The string to ahead progress bar during the run when `progress` is
            True. The default is 'Fitting'.

        Returns
        -------
        result : dict
            The simulation and fitting result.
        """
        n_parallel = get_parallel_number(n_parallel)

        init_params = jax.tree.map(jnp.array, init_params)
        assert set(init_params) == set(params_names)

        # check if all params shapes are the same
        param_shapes = [np.shape(v)[:-1] for v in init_params.values()]
        assert all(i == param_shapes[0] for i in param_shapes)

        # the data shape is (nsim, nchan)
        data_shapes = [np.shape(data[k])[:-1] for k in ndata if k != 'total']
        assert all(i == data_shapes[0] for i in data_shapes)
        assert all(len(i) == 1 for i in data_shapes)
        nsim = data_shapes[0][0]

        # get initial parameters arrays in unconstrained space,
        init = jnp.array([init_params[k] for k in params_names]).T
        assert init.ndim <= 2
        if init.ndim == 2:
            assert init.shape[0] == nsim
        if init.ndim == 1:
            init = jnp.full((nsim, len(init)), init)
        init = jax.vmap(constr_arr_to_unconstr_arr)(init)

        # fit result container
        result = {
            'data': data,
            'params': {k: jnp.empty(nsim) for k in params_names},
            'models': {
                i: jnp.empty((nsim, ndata[k]))
                for k, v in data_group.items()
                for i in [k, *map('{}_model'.format, v)]
            },
            'deviance': {
                'total': jnp.empty(nsim),
                'group': {k: jnp.empty(nsim) for k in data_group},
                'point': {k: jnp.empty((nsim, ndata[k])) for k in data_group},
            },
            'valid': jnp.full(nsim, True, bool),
        }

        # fit simulation data
        if parallel:
            result = fit_in_parallel(
                result, init, run_str, progress, update_rate, n_parallel
            )
        else:
            result = fit_in_sequence(
                result, init, run_str, progress, update_rate
            )

        return result

    def simulate_and_fit(
        seed: int,
        free_params: dict[ParamName, JAXArray],
        model_values: dict[str, JAXArray],
        n: int = 1,
        parallel: bool = True,
        n_parallel: int | None = None,
        progress: bool = True,
        update_rate: int = 50,
        run_str: str = 'Fitting',
    ) -> dict:
        """Simulate data and then fit the simulation data.

        Parameters
        ----------
        seed : int
            The random number generator seed used for data simulation.
        free_params : dict
            The free parameters values in unconstrained space.
        model_values : dict
            The model values corresponding to `free_params`.
        n : int, optional
            The number of simulations of each model value, by default 1.
        parallel : bool, optional
            Whether to fit in parallel, by default True.
        n_parallel : int, optional
            The number of parallel processes when `parallel` is ``True``.
            Defaults to ``jax.local_device_count()``.
        progress : bool, optional
            Whether to show progress bar, by default True.
        update_rate : int, optional
            The update rate of the progress bar, by default 50.
        run_str : str, optional
            The string to ahead progress bar during the run when `progress` is
            True. The default is 'Fitting'.

        Returns
        -------
        result : dict
            The simulation and fitting result.
        """
        model_values = {
            f'{k}_model': model_values[f'{k}_model'] for k in simulators
        }

        # TODO: support posterior prediction with n > 1
        # check if all model shapes are the same
        shapes = [
            np.shape(model_values[f'{k}_Non_model'])[:-1]
            for k in ndata
            if k != 'total'
        ]
        assert all(i == shapes[0] for i in shapes)
        assert not (shapes[0] != () and n > 1)

        # simulate data
        sim_data = simulate(int(seed), model_values, int(n))

        # fit simulation data
        return batch_fit(
            free_params,
            sim_data,
            parallel,
            n_parallel,
            progress,
            update_rate,
            run_str,
        )

    # =============== functions used in simulation procedure ==================

    return Helper(
        ndata=ndata,
        nparam=nparam,
        dof=dof,
        data_names=list(data.keys()),
        statistic=stat,
        channels=channels,
        obs_data=obs_data,
        data=dict(data),
        model=dict(model),
        seed=rng_seed,
        sampling_dist=sampling_dist,
        numpyro_model=numpyro_model,
        params_names={
            'free': free_names,
            'deterministic': deterministic_names,
            'interest': interest_names,
            'all': params_names,
        },
        params_default=unconstr_dic_to_params_dic(
            dict(zip(free_names, default_unconstr_arr, strict=True))
        ),
        params_setup=model_info.setup,
        params_latex=pname_to_latex,
        params_unit=pname_to_unit,
        params_log=pname_to_log,
        params_comp_latex=pname_to_comp_latex,
        free_default=free_default,
        get_sites=get_sites,
        get_params=get_params,
        get_models=get_models,
        get_loglike=get_loglike,
        get_mle=get_mle,
        params_covar=params_covar,
        deviance_total=deviance_total,
        deviance=deviance,
        residual=residual,
        constr_arr_to_unconstr_arr=constr_arr_to_unconstr_arr,
        constr_dic_to_unconstr_arr=constr_dic_to_unconstr_arr,
        unconstr_dic_to_params_dic=unconstr_dic_to_params_dic,
        simulate=simulate,
        simulate_and_fit=simulate_and_fit,
        batch_fit=batch_fit,
    )


class Helper(NamedTuple):
    """Helper for fitting and analysis."""

    ndata: dict[str, int]
    """The number of channels in each dataset and the total number of channels.
    """

    nparam: int
    """The number of free parameters in the model."""

    dof: int
    """The degree of freedom."""

    data_names: list[str]
    """Name of each data."""

    statistic: dict[str, Statistic]
    """The statistic used in each dataset."""

    channels: dict[str, np.ndarray]
    """Channel information of the datasets."""

    obs_data: dict[str, JAXArray]
    """The datasets of observations, including net counts, counts in the "on"
    and "off" measurements.
    """

    data: dict[str, FixedData]
    """FixedData instances."""

    model: dict[str, CompiledModel]
    """Compiled spectral models."""

    seed: dict[str, int]
    """Random number generator seed."""

    sampling_dist: dict[str, tuple[Literal['norm', 'poisson'], tuple]]
    """Sampling distribution of observation data, this is used for probability
    integral transform calculation.
    """

    numpyro_model: Callable[[bool], None]
    """The numpyro model for spectral fitting."""

    params_names: dict
    """The names of parameters in the model."""

    params_default: dict[str, JAXFloat]
    """The default values of parameters."""

    free_default: dict[str, dict[ParamName, JAXFloat] | JAXArray]
    """The default values of free parameters."""

    params_setup: dict[ParamName, tuple[ParamName, ParamSetup]]
    """The mapping from forwarded parameters names to parameters names."""

    params_latex: dict[ParamName, str]
    """The LaTeX representation of parameters."""

    params_unit: dict[ParamName, str]
    """The unit of parameters."""

    params_log: dict[ParamName, bool]
    """Whether the parameters are in log space."""

    params_comp_latex: dict[ParamName, str]
    """The LaTeX representation of parameter's component."""

    get_sites: Callable[
        [JAXArray],
        dict[
            Literal['params', 'models', 'loglike'],
            dict[str, JAXFloat | JAXArray],
        ],
    ]
    """Get parameters in constrained space, models values and log likelihood,
    given free parameters array in unconstrained space.
    """

    get_params: Callable[[Mapping], dict]
    """Get parameters' values in constrained space given numpyro model sites.
    """

    get_models: Callable[[Mapping], dict]
    """Get model values given numpyro model sites."""

    get_loglike: Callable[[Mapping], dict]
    """Get log likelihood given numpyro model sites."""

    get_mle: Callable[[JAXArray], tuple[JAXArray, JAXArray]]
    """Get the MLE and error of all parameters in constrained space,
    given MLE of free parameters in unconstrained space.
    """

    params_covar: Callable[[JAXArray, JAXArray], JAXArray]
    """Calculate covariance matrix of all parameters in constrained space,
    given values and covariance matrix of free parameters in unconstrained
    space.
    """

    deviance_total: Callable[[JAXArray], JAXFloat]
    """Calculate total deviance given free parameters array in unconstrained
    space.
    """

    deviance: Callable[[JAXArray], dict[str, JAXArray]]
    """Calculate total, group and point deviance given free parameters array in
    unconstrained space.
    """

    residual: Callable[[JAXArray], JAXArray]
    """Calculate deviance residual (i.e., sqrt deviance) given free parameters
    array in unconstrained space.
    """

    constr_arr_to_unconstr_arr: Callable[[JAXArray], JAXArray]
    """Covert free parameters array from constrained space into
    unconstrained space.
    """

    constr_dic_to_unconstr_arr: Callable[[ParamNameValMapping], JAXArray]
    """Covert free parameters dict from constrained space to array in
    unconstrained space.
    """

    unconstr_dic_to_params_dic: Callable[
        [ParamNameValMapping], ParamNameValMapping
    ]
    """Get parameters dict in constrained space, given a free parameters dict
    in unconstrained space.
    """

    simulate: Callable[[int, dict[str, JAXArray], int], dict[str, JAXArray]]
    """Function to simulate data."""

    simulate_and_fit: Callable[
        [int, dict, dict, int, bool, int, bool, int, str], dict
    ]
    """Function to simulate data and then fit the simulation data."""

    batch_fit: Callable[
        [dict[str, JAXArray], dict[str, JAXArray], bool, int, bool, int, str],
        dict,
    ]
    """Function to fit simulation data."""



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
import multiprocess as mp
import nautilus
import nautilus.pool as nautilus_pool

from elisa.infer.samplers.util import uniform_reparam_model

if TYPE_CHECKING:
    from collections.abc import Callable

    from numpy.typing import NDArray


class NautilusSampler:
    def __init__(
        self,
        numpyro_model: Callable,
        model_args: tuple = (),
        model_kwargs: dict | None = None,
        seed: int = 42,
        ignore_nan: bool = False,
        **kwargs: dict,
    ):
        if ignore_nan:
            warnings.warn(
                'setting `ignore_nan` to True may fail to spot potential '
                'issues of the model',
                Warning,
            )

        self._model_info = mi = uniform_reparam_model(
            numpyro_model,
            model_args,
            model_kwargs,
            rng_seed=seed,
        )

        @jax.jit
        def log_prob_fn(cube_and_derived):
            log_p = mi.log_prob_fn(mi.unravel(cube_and_derived[: mi.ndim]))
            if ignore_nan:
                log_p = jnp.nan_to_num(log_p, nan=-1e300)
            return log_p

        if 'pool' in kwargs:
            kwargs['vectorized'] = False
            old_method = mp.get_start_method()
            if old_method != 'spawn':
                mp.set_start_method('spawn', force=True)
            else:
                old_method = ''
            # monkey patching the pool for compatibility with JAX
            old_pool = nautilus_pool.Pool
            nautilus_pool.Pool = mp.Pool
        else:
            kwargs['vectorized'] = True
            log_prob_fn = jax.jit(jax.vmap(log_prob_fn))
            old_method = ''
            old_pool = None

        self._sampler = nautilus.Sampler(
            prior=lambda x: x,
            likelihood=lambda x: jax.device_get(log_prob_fn(x)),
            n_dim=mi.ndim,
            pass_dict=False,
            seed=seed,
            **kwargs,
        )

        if old_method:
            mp.set_start_method(old_method, force=True)

        if old_pool is not None:
            nautilus_pool.Pool = old_pool

    def run(self, **kwargs) -> dict[str, NDArray[float]]:
        kwargs.setdefault('verbose', True)
        kwargs['discard_exploration'] = True
        sampler = self._sampler
        success = sampler.run(**kwargs)
        if success:
            u_samples, *_ = sampler.posterior(
                return_as_dict=False,
                equal_weight=True,
            )
            u_samples = jax.vmap(self._model_info.unravel)(u_samples)
            samples = jax.vmap(self._model_info.postprocess_fn)(u_samples)
            samples = jax.device_get(samples)
            return samples
        else:
            raise RuntimeError(
                'Sampling failed due to limits were reached, please set a '
                'larger `n_like_max` or `timeout`. You can also resume the '
                'sampler from previous one, providing `filepath` and `resume`.'
            )

    @property
    def ess(self) -> int:
        return int(self._sampler.n_eff)

    @property
    def lnZ(self) -> float | None:
        return self._sampler.log_z
